/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <string.h>
#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"
#include "rfc822.h"
#include "keymap.h"

static ALIAS **adrtbl = 0;

/*
 * This is the callback routine from mutt_menuLoop() which is used to generate
 * a menu entry for the requested item number.
 */

void alias_makeEntry (char *s, int num) {
  sprintf(s, "%2d   %-10s   =>   ", num+1, adrtbl[num]->name);
  rfc822_write_address(s, adrtbl[num]->addr);
}

/*
 * Displays a menu of the user's aliases.  Returns a pointer to the alias
 * the user selects.
 */

ADDRESS *alias_menu (void) {
  ALIAS *aliasp;
  ADDRESS *padr = 0;
  MUTTMENU *menu;
  int i, adrmax, done = 0;

  if (!Aliases) {
    mutt_error("You have no aliases!");
    return 0;
  }

  for (adrmax = 0, aliasp = Aliases; aliasp; aliasp = aliasp->next, adrmax++);
  adrtbl = (ALIAS **)safe_malloc(sizeof(ALIAS *) * (adrmax+1));
  adrtbl[adrmax] = 0;
  for (i=0, aliasp = Aliases; aliasp; aliasp = aliasp->next, i++)
    adrtbl[i] = aliasp;

  menu = mutt_newMenu();
  menu->prompt = safe_strdup("Alias: ");
  menu->title = safe_strdup("Alias Menu");
  menu->max = adrmax;
  menu->makeEntry = alias_makeEntry;
  menu->menu = MENU_ALIAS;
#ifdef NCURSES_MOUSE_VERSION
  if (Mousemask) {
    menu->tool = Generic_toolbar;
    toolbar_init(Generic_toolbar);
  }
#endif /* NCURSE_MOUSE_VERSION */

  while (!done) {
    switch (mutt_menuLoop(menu)) {
    case OP_GENERIC_SELECT_ENTRY:
      padr = adrtbl[menu->current]->addr;
      done = 1;
      break;

    case OP_GENERIC_EXIT:
      done = 1;
      break;
    }
  }

  mutt_menuDestroy(&menu);
  safe_free((void **)&adrtbl);
  return(padr);
}
