/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <pwd.h>
#include <string.h>
#include "mutt.h"
#include "mutt_curses.h"
#include "rfc822.h"

static ADDRESS *lookup_alias (char *s)
{
  ALIAS *t = Aliases;

  while (t) {
    if (strcmp (s, t->name) == 0) return (t->addr);
    t = t->next;
  }
  return (0);
}

ADDRESS *mutt_expand_aliases (ADDRESS *a)
{
  ADDRESS *head = 0, *last = 0, *t, *w;

  while (a) {
    if (a->host && *a->host == '@') {
      t = lookup_alias(a->mailbox);

      /* try to prevent bonehead alias loops... */
      if (t && strcmp(t->mailbox, a->mailbox) == 0 && strcmp(t->host, a->host) == 0)
	t=0;

      if (!t) {
        safe_free((void **)&a->host);
        a->host = safe_strdup(Fqdn);
	
	if (!a->personal) {
	  struct passwd *pw = getpwnam(a->mailbox);
	  char buffer[256], *p;

	  if (pw) {
	    strcpy(buffer, pw->pw_gecos);
	    if ((p = strchr(buffer, ','))) *p = 0;
	    a->personal = safe_strdup(buffer);
	  }
	}

        if (last) {
          last->next = a;
          last = last->next;
          a = a->next;
          last->next = 0;
        }
        else {
          head = last = a;
          a = a->next;
          last->next = 0;
        }
        continue;
      }
      else {
	w = rfc822_cpy_adr (t);
	w = mutt_expand_aliases (w);
	if (head)
	  last->next = w;
	else
	  head = last = w;
	while (last->next) last = last->next;
	t = a;
	a = a->next;
	t->next = 0;
	mutt_free_address (&t);
      }
    }
    else {
      if (head) {
	last->next = a;
	last = last->next;
	a = a->next;
	last->next = 0;
      }
      else {
	head = last = a;
	a = a->next;
	head->next = 0;
      }
    }
  }
  return (head);
}

void mkalias (void)
{
  HEADER *cur = Context->hdrs[*Context->v2r[Context->current]];
  ALIAS *new, *t;
  char buffer[STRING];
  FILE *rc;

  move(LINES-3, 0);

  strcpy (buffer, cur->reply_to ? cur->reply_to->mailbox : cur->from->mailbox);
  ci_get_field ("Alias as: ", buffer, 0, 0);
  if (buffer[0] == 0) return;
  new = (ALIAS *) safe_malloc (sizeof (ALIAS));
  new->name = safe_strdup (buffer);
  new->next = 0;
  new->addr = 0;

  if (cur->reply_to)
    sprintf (buffer, "%s@%s", cur->reply_to->mailbox, cur->reply_to->host);
  else
    sprintf (buffer, "%s@%s", cur->from->mailbox, cur->from->host);
  ci_get_field ("Address: ", buffer, 0, 0);
  if (buffer[0] == 0) {
    safe_free ((void **) &new->name);
    safe_free ((void **) &new);
    return;
  }
  rfc822_parse_adrlist (&new->addr, buffer, Fqdn);

  if (cur->reply_to && cur->reply_to->personal)
    strcpy (buffer, cur->reply_to->personal);
  else if (cur->from->personal)
    strcpy (buffer, cur->from->personal);
  else
    buffer[0] = 0;
  ci_get_field ("Personal name: ", buffer, 0, 0);
  new->addr->personal = safe_strdup (buffer);

  sprintf (buffer, "[%s = ", new->name);
  rfc822_write_address (buffer, new->addr);
  strcpy (buffer+strlen (buffer), "]  Accept?");
  if (mutt_yesorno(buffer, 1) == 0) {
    safe_free ((void **) &new->name);
    mutt_free_address(&new->addr);
    safe_free ((void **) &new);
    return;
  }

  if ((t = Aliases)) {
    while (t->next) t = t->next;
    t->next = new;
  }
  else
    Aliases = new;

  sprintf (buffer, "%s/.muttrc", Homedir);
  rc = fopen (buffer, "a");
  buffer[0] = 0;
  rfc822_write_address (buffer, new->addr);
  fprintf (rc, "alias %s %s\n", new->name, buffer);
  fclose (rc);

  mutt_error ("Alias added.");
}
