#include "mutt.h"
#include <string.h>
#include <ctype.h>
#include "keymap.h"

extern char Init_error[SHORT_STRING];

static char *strtokq (char *s, char *tokens)
{
  static char *p;
  char quote_char = 0, *r, *t;
  
  if (s) {
    /*
     * initial call.
     */
    p = s;
  }

  if (!p) return 0;

  if (*p == '\'' || *p == '\"') {
    quote_char = *p;
    p++;
    r = p;
    while (*p && *p != quote_char) p++;
    *p = 0;
    p++;
    while (strchr(tokens, *p) != 0) p++;
    return(r);
  }
  r = p;
  if ((t = strpbrk(p, tokens))) {
    *t++ = 0;
    while (*t && strchr(tokens, *t) != 0) t++;
    p = t;
    return(r);
  }
  else
    p = 0; /* all done */
  return(r);
}

/*
 * bind '<key_sequence>' function-name menu-name
 */
int mutt_parse_bind (char *s)
{
  int count = 0, control, menu = -1, i;
  char key = -1, *op = 0;
  struct binding_t *bindings = 0;

  while ((s = strtokq(s, " \t"))) {
    switch (count) {

    case 0:

      control = 0;
      if (s[0] == '\\' && (s[1] == 'c' || s[1] == 'C')) {
        control = 1;
        s += 2;
      }
      if (control) {
        *s = toupper(*s);
        key = *s - 'A' - 1;
      }
      else if (strcmp(s, "\\n") == 0)
        key = '\n';
      else if (strcmp(s, "\\r") == 0)
        key = '\r';
      else
        key = *s;
      break;

    case 1:

      op = s;
      break;

    case 2:

      if (strcmp("alias-menu", s) == 0) {
	bindings = 0;
	menu = MENU_ALIAS;
      }
      else if (strcmp("attach-menu", s) == 0) {
        bindings = Op_attach;
        menu = MENU_ATTACH;
      }
      else if (strcmp("folder-menu", s) == 0) {
        bindings = 0;
        menu = MENU_FOLDER;
      }
      else if (strcmp("internal-pager", s) == 0) {
        bindings = Op_pager;
        menu = MENU_PAGER;
      }
      else if (strcmp("main-menu", s) == 0) {
        bindings = Op_main;
        menu = MENU_MAIN;
      }
      else if (strcmp("send-menu", s) == 0) {
	bindings = Op_send;
	menu = MENU_SEND;
      }
      else {
        sprintf(Init_error, "%s: no such menu", s);
        return(-1);
      }
    }

    count++;
    s = 0;
  }

  /*
   * Check to make sure that we got enough arguments to proceed.
   */
  if (key == -1 || !op || menu == -1) {
    strcpy(Init_error, "bind: too few arguments.");
    return(-1);
  }

  if (menu == MENU_PAGER || menu == MENU_MAIN || menu == MENU_SEND) {
    /*
     * These menus have keymaps of their own and do not refer to the
     * generic map.
     */
    for (i = 0; bindings[i].name; i++)
      if (strcmp(op, bindings[i].name) == 0) {
	keymap_bind(key, menu, bindings[i].op);
	return(0);
      }
    
  }
  else {
    /*
     * First check the "generic" list of commands.
     */
    for (i=0; Op_generic[i].name; i++) {
      if (strcmp(op, Op_generic[i].name) == 0) {
	keymap_bind(key, menu, Op_generic[i].op);
	return(0);
      }
    }

    /*
     * Now check the menu-specific list of commands (if they exist).
     */
    if (bindings)
      for (i=0; bindings[i].name; i++)
	if (strcmp(op, bindings[i].name) == 0) {
	  keymap_bind(key, menu, bindings[i].op);
	  return(0);
	}
  }

  sprintf(Init_error, "%s: no such function in map", op);
  return(-1);
}
