/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <config.h> /* needed for HAVE_SYS_PARAM_H */
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h> /* needed for definition of BSD on BSD-ish systems. */
#endif /* HAVE_SYS_PARAM_H */
#include <termios.h>
#if defined(BSD) && (BSD >= 199103)
#include <sys/ioctl_compat.h>
#endif /* BSD */
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
/* SunOS 4.x gets redefinition errors, but Linux needs this. */
#if !defined(__sun__)
#include <sys/ioctl.h>
#endif
#include <string.h>
#include "mutt.h"
#include "mutt_curses.h"

int ci_any_key_to_continue (const char *s)
{
  struct termios t;
  int f, ch;

  f = open("/dev/tty", O_RDONLY);
#if (defined(BSD) && (BSD >= 199103)) || defined(_HPUX_SOURCE) || defined(__convex__)
  tcgetattr(f, &t);
#else
  ioctl(f, TCGETS, &t);
#endif
  t.c_lflag &= ~(ICANON | ECHO);
  t.c_cc[VMIN] = 1;
  t.c_cc[VTIME] = 0;
#if (defined(BSD) && (BSD >= 199103)) || defined(_HPUX_SOURCE) || defined(__convex__)
  tcsetattr(f, TCSADRAIN, &t);
#else
  ioctl(f, TCSETS, &t);
#endif
  if (s) fputs(s, stdout);
  else fputs("Press any key to continue...", stdout);
  fflush(stdout);
  ch = fgetc(stdin);
  fflush(stdin);
  t.c_cflag |= (ICANON | ECHO);
#if (defined(BSD) && (BSD >= 199103)) || defined(_HPUX_SOURCE) || defined(__convex__)
  tcsetattr(f, TCSADRAIN, &t);
#else
  ioctl(f, TCSETS, &t);
#endif
  close(f);
  fputc('\n', stdout);
  return(ch);
}

/*
**  For all the 8-bit charsets
*/
int IsPrint(int c) 
{
    return (isprint(c) || (c > 0xA0 && c < 0xFF));
}

/*
**  Display a 8-bit character as is in using 'ncurses' library.
**    - no problem in using 'curses' library
*/
int addch8(chtype c)
{
#if defined(HAVE_NCURSES_H) || defined(HAVE_NCURSES_LIB)
    return (c & 0x80)? addch((c & 0xFF) | attr_get()): addch(c);
#else
    return addch(c);
#endif
}

/*
**  Allow all the 8-bit charsets as an input beside 7-bit US-ASCII charset.
**  - European charset(ISO-8859-X) uses 8-bit single-byte coded characters
**  - CJK(Chinese-Japan-Korean) charset uses 8-bit double-byte coded characters
*/
int ci_enter_string (char *buffer, int y, int x, int flags, int initch)
{
  int ch;
  short lastchar, i, len, j, beg = 0;
  int pass = flags & CI_password;
  int complete = flags & CI_complete;
  char ch8bit;
  int iso8859 = (!strncasecmp(Charset, "ISO-8859-", 9))? 1: 0;

  len = COLS - x - 1;

redraw_it:
  move(y,x);
  lastchar = i = strlen(buffer);
  if (!pass && i>0) {
    beg = lastchar - len;
    if (beg < 0)
      beg = 0;
    else if (beg > 0)
      clrtoeol();
    for (j=beg; j<lastchar && j<beg+len; j++) {
      addch8(buffer[j]);
    }
  }
  refresh();
  FOREVER {
    if (!initch)
      ch = ci_getch();
    else {
      ch = initch;
      initch = 0;
    }
    dprint(5, (debugfile, "ci_enter_string(): ch = %d.\n", ch));
    if (complete && (ch == ' ' || ch == '\t')) {
      buffer[i]=0;
      mutt_complete(buffer);
      goto redraw_it;
    }
    else if (ch == KEY_BACKSPACE || ch == 127) {
      if (i == 0) {
        beep();
        refresh();
        continue;
      }
      if (i == lastchar) {
        lastchar--;
        i--;
        if(!iso8859)
          if((buffer[i] & 0x80) & (buffer[i-1] & 0x80)) { lastchar--; i --; }
		if(i<beg) i=beg; if(lastchar<beg) lastchar=beg;
        if (!pass) {
          if (i < beg) {
            move(y,x);
            clrtoeol();
            beg -= len;
            if (beg < 0) beg = 0;
            for(j=beg; j<lastchar && j<beg+len; j++) {
              addch8(buffer[j]);
            }
          }
          else {
            move(y,x+i-beg); clrtoeol();
            if(!iso8859) 
              if((buffer[i-1] & 0x80) & (buffer[i-2] & 0x80)) delch();
            if(i-beg>0) delch();
          }
        }
      }
      else {
        if(!iso8859)
          ch8bit = ((buffer[i-1] & 0x80) & (buffer[i-2] & 0x80))? 1: 0; 
        else ch8bit = 0;
        for (j=i; j<lastchar; j++) {
          if(ch8bit) buffer[j-2] = buffer[j];
          else buffer[j-1] = buffer[j];
        }
        i--;
        lastchar--;
        if(ch8bit) { lastchar--; i--; }
        if(i<beg) i=beg; if(lastchar<beg) lastchar=beg;
        if (!pass) {
          buffer[lastchar] = 0;
          move(y,x+i-beg);
          clrtoeol();   
          for(j=i; j<lastchar && j<beg+len; j++) {
            addch8(buffer[j]);
          }
          move(y,x+i-beg);
        }
      }
    }
    else if (ch == ctrl('A') || ch == KEY_HOME) {
      i = 0;
      if (!pass) {
        move(y,x);
        beg = 0;
        if (x + lastchar > len) {
          clrtoeol();
          for(j=0;j<len;j++) {
            addch8(buffer[j]);
          }
          move(y,x);
        }
      }
    }
    else if (ch == ctrl('E') || ch == KEY_END) {
      if (!pass) {
        if (lastchar < beg + len)
          move(y,x+lastchar-beg);
        else {
          move(y,x);
          clrtoeol();
          beg = lastchar-len/2;
          for(j=beg; j<lastchar; j++) {
            addch8(buffer[j]);
          }
        }
      }
      i = lastchar;
    }
    else if (ch == ctrl('U')) {
      lastchar = i = 0;
      buffer[0] = 0;
      if (!pass) {
        move(y,x);
        clrtoeol();
        beg = 0;
      }
    }
    else if (ch == ctrl('K')) {
      lastchar = i;
      buffer[i] = 0;
      if (!pass) {
        clrtoeol();
      }
    }
    else if (ch == ctrl('B') || ch == KEY_LEFT) {
      if (i == 0)
        beep();
      else {
        i--;
        if(!iso8859) 
          if((buffer[i] & 0x80) & (buffer[i-1] & 0x80)) i--;
        if (!pass) {
          if (i < beg) {
            beg -= len;
            if (beg < 0) beg = 0;
            move(y,x);
            clrtoeol();
            for(j=beg;j<beg+len;j++) {
              addch8(buffer[j]);
            }
          }
          move(y,x+i-beg);
        }
      }
    }
    else if (ch == ctrl('F') || ch == KEY_RIGHT) {
      if (i == lastchar)
        beep();
      else {
        if(!iso8859) 
          if((buffer[i] & 0x80) & (buffer[i+1] & 0x80)) i++;
        i++;
        if (!pass) {
          if (i >= beg + len) {
            beg = i;
            move(y,x);
            clrtoeol();
            for(j=beg; j<lastchar && j<beg+len; j++) {
              addch8(buffer[j]);
            }
            move(y,x);
          }
          else
            move(y,x+i-beg);
        }
      }
    }
    else if (ch == ctrl('D')) {
      for (j=i;j<lastchar;j++)
        buffer[j] = buffer[j+1];
      lastchar--;
      if (!pass) {
        clrtoeol();
        for (j=i;j<lastchar && j<beg+len; j++)
          addch8(buffer[j]);
        move(y,x+i-beg);
      }
    }
    else if (ch == ctrl('G')) { /* ABORT */
      buffer[0] = 0;
      return(-1);
    }
    else if (CI_is_return(ch)) {
      buffer[lastchar] = 0;
      return(0);
    }
    else if (IsPrint(ch)) {
      if(!iso8859) {
        if(ch & 0x80) initch = ci_getch();
        ch8bit = (initch && IsPrint(initch))? 1: 0;
      }
      else ch8bit = 0;
      if (i == lastchar) {
        buffer[i++] = ch;
        lastchar++;
        if(ch8bit) { buffer[i++] = initch; lastchar++; }
        if (!pass) {
          if (i > beg+len) {
            beg = i - len/2;
            move(y,x);
            clrtoeol();
            for (j=beg;j<lastchar;j++) {
              addch8(buffer[j]);
            }
          }
          else {
            addch8(ch);
            if(ch8bit) addch8(initch);
          }
        }
      }
      else {
        for (j=lastchar; j>i; j--) {
          if(ch8bit) buffer[j+1] = buffer[j-1];
          else buffer[j] = buffer[j-1];
        }
        buffer[i] = ch;
        i++;
        lastchar++;
        if(ch8bit) { buffer[i++] = initch; lastchar++; }
        if (!pass) {
          if (i > beg + len) {
            move(y,x);
            clrtoeol();
            beg = (ch8bit)? i-2: i-1;
            for(j=beg;j<lastchar && j<beg+len; j++) {
              addch8(buffer[j]);
            }
          }
          else {
            clrtoeol();
            for(j=(ch8bit)? i-2: i-1;j<lastchar && j<beg+len; j++) {
              addch8(buffer[j]);
            }
          }
          move(y,x+i-beg);
        }
      }
      if(ch8bit) initch = 0;
    }
    else
      beep();
    refresh();
  }
}
     
int ci_get_field (const char *field, char *buffer, int initch, int complete)
{
  int ret;
  int len = strlen(field); /* in case field==buffer */

  ClearLine(LINES-2);
  addstr(field);
  refresh();
  ret = ci_enter_string(buffer, LINES-2, len, complete, initch);
  ClearLine(LINES-2);
  return(ret);
}

void ci_clear_error (void)
{
  if (Errorbuf[0]) {
    Errorbuf[0] = 0;
    ClearLine(LINES-1);
    refresh();
  }
}

void ci_edit_file (const char *p)
{
  char cmd[256];

  endwin();
  sprintf(cmd, "%s %s", Editor, p);
  mutt_system(cmd);
}

int mutt_yesorno (const char *msg, int def)
{
  int ch;

  move(LINES-2, 0);
  clrtoeol();
  printw("%s: [%c] ", msg, def ? 'y' : 'n');
  FOREVER {
    refresh();
    ch = ci_getch();
    if (CI_is_return(ch)) break;
    else if (ch == 'y') {
      def = 1;
      break;
    }
    else if (ch == 'n') {
      def = 0;
      break;
    }
    else
      beep();
  }
  addstr(def ? "Yes" : "No");
  refresh();
  return def;
}

int ci_getch (void)
{
  if (setjmp(Jumpbuf)) {
    /* ci_sigwinch is set to true, so caller can repaint the screen. */
    fflush(stdin);
    return(ERR);
  }
  return(getch());
}
