/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <string.h>
#include <setjmp.h>
#include "mutt.h"
#include "mutt_curses.h"
#include "rfc822.h"
#include "toolbar.h"
#include "keymap.h"

#define msgno(X) *Context->v2r[X]

#define REDRAW_FULL 1
#define REDRAW_INDEX 2
#define REDRAW_CURRENT 3
#define REDRAW_MOTION 4

#define HDR_OFFSET 3
#define HDR_CMD (LINES-2)
#define HDR_LEN (HDR_CMD-HDR_OFFSET-2)

#define CI_generate_header(A,B,C) ci_generate_header(A,B,C)
#define CI_clearline(x) move(x,0); clrtoeol()

int ci_sigwinch = 0;
jmp_buf Jumpbuf;
/* The folder the user last saved to.  Used by ci_save_message() */
static char ci_last_folder[_POSIX_PATH_MAX];

#ifdef NCURSES_MOUSE_VERSION
TOOLBAR Main_toolbar[] = {
  { "Quit",  OP_MAIN_QUIT, 0 },
  { "Top",   OP_MAIN_FIRST_MESSAGE, 0 },
  { "Bot",   OP_MAIN_LAST_MESSAGE, 0 },
  { "NxtPg", OP_MAIN_NEXT_PAGE, 0 },
  { "PrvPg", OP_MAIN_PREVIOUS_PAGE, 0 },
  { "Del",   OP_MAIN_DELETE_MESSAGE, 0 },
  { "UnDel", OP_MAIN_UNDELETE_MESSAGE, 0 },
  { "Attch", OP_MAIN_VIEW_ATTACHMENTS, 0 },
  { "Mail",  OP_MAIN_MAIL, 0 },
  { "Reply", OP_MAIN_REPLY, 0 },
  { "Group", OP_MAIN_GROUP_REPLY, 0 },
  { "Forw",  OP_MAIN_FORWARD_MESSAGE, 0 },
  { 0, 0, 0 }
};
#endif

static void ci_generate_header (char *s, HEADER *h, int l)
{
  int len;
  int cols = option(OPTARROWCURSOR) ? COLS-2 : COLS;

  mutt_generate_header(s, h, l);
  len = strlen(s);
  if (len < cols)
    sprintf(s+len, "%*.*s", cols-len, cols-len, "");
  else
    s[cols] = 0;
}

static void ci_show_last_error (void)
{
  if (Errorbuf[0]) {
    mvaddstr(LINES-1, (COLS-strlen(Errorbuf))/2, Errorbuf);
    refresh();
  }
}

#ifdef METAMAIL
static int ci_invoke_metamail (HEADER *h, FILE *f)
{
  long chunk, bytes = h->length;
  FILE *fout;
  char *tempfile, buffer[LONG_STRING];

  tempfile = mutt_new_tempfile();
  if ((fout = fopen(tempfile, "w")) == NULL) return -1;
  if (fseek(f, h->offset, 0) == -1) {
    fclose(fout);
    safe_free((void **)&tempfile);
    return -1;
  }
  while (bytes > 0) {
    chunk = (bytes > sizeof(buffer)) ? sizeof(buffer) : bytes;
    chunk = fread(buffer, 1, chunk, f);
    fwrite(buffer, 1, chunk, fout);
    bytes -= chunk;
  }
  fclose(fout);
  sprintf(buffer, "%s -P -z %s", METAMAIL, tempfile);
  mutt_system(buffer);
  safe_free((void **)&tempfile);
  return 0;
}
#endif /* METAMAIL */

/*
**  - modified by pms@romance.kaist.ac.kr
*/ 
static int ci_display_message (void)
{
  extern void DecodeHeader(char*, const char*);
  HEADER *cur = Context->hdrs[msgno(Context->current)];
  char tempfile[_POSIX_PATH_MAX], buffer[STRING], ts[STRING];
  STATE *st;
  int rc, builtin=0;

  if (!cur->read) {
    cur->read = 1;    /* it is read now */
    cur->changed = 1; /* mark that this message has changed */
  }

#ifdef METAMAIL
  if (cur->metamail) {
    ci_invoke_metamail(cur, Context->folder->fp);
    return 0;
  }
#endif

  mutt_parse_mime_message(Context->hdrs[*Context->v2r[Context->current]]);

  st = mutt_new_state();
  st->displaying = 1;

/* PMS - to solve a displaying problem of PGP-message( To-Do )
       - Use builtin Pager if you are using PGP mail
  if (strcmp(Pager, "builtin") == 0 || (cur->content->subtype != NULL &&
      strcasecmp("encrypted", cur->content->subtype) == 0)) {
*/
  if (strcmp(Pager, "builtin") == 0) {
    sprintf(tempfile, "%s/mutt.XXXXXX", Tempdir);
    mktemp(tempfile);
    st->fpout = fopen(tempfile, "w");
    builtin++;
  }
  else {
    endwin();
    mutt_create_filter(Pager, &(st->fpout), 0);
  }

  sprintf(buffer, "%s %d/%d: ", cur->deleted ? "[Deleted]" : "Message",
	  msgno(Context->current) + 1,
	  Context->msgcount);
  if (cur->from && cur->from->personal)
    sprintf(buffer+strlen(buffer), "%-20.20s", cur->from->personal);
  else
    rfc822_write_address(buffer+strlen(buffer), cur->from);
  DecodeHeader(ts, buffer);
  fprintf(st->fpout, "%s\n", ts);

  if (cur->subject) {
    int len = strlen (cur->subject);

    len = (COLS-len)/2;
    if (len < 0) {
      /* The subject is longer than the current line! */
      fprintf (st->fpout, "%*.*s\n", COLS, COLS, cur->subject);
    }
    else
      fprintf (st->fpout, "%*.*s%s\n", len, len, "", cur->subject);
  }
  fputc ('\n', st->fpout);

  st->fpin = Context->folder->fp;
  mutt_copy_header(cur, st->fpout, NIL);
  state_putc('\n', st);
  mutt_body_handler(cur->content, st);
  fclose(st->fpout);
  mutt_free_state(&st);

  if (builtin) {
    /* Invoke the builtin pager */
    rc = ci_simple_pager(tempfile); /* removes file upon completion */
  }
  else {
    wait (&rc);
    keypad(stdscr, TRUE);
    if (option(OPTPROMPTAFTER)) {
      rc = ci_any_key_to_continue("Command: ");
      rc = keymap_lookup(rc, MENU_PAGER);
    }
    else
      rc = 0;
  }
  return(rc);
}

/*
** Support reccurent messagebar - pms@romance.kaist.ac.kr 
*/
static int ci_next_undeleted (void)
{
  short i, j;

  i = (Context->current < (Context->vcount-1))? Context->current+1: 0;
  for(j=1; Context->hdrs[msgno(i)]->deleted; i++,j++) {
    if(i==Context->vcount-1) i = -1;
    if(j==Context->vcount-1) break;
  }
  return(i);
}

/*
** Support reccurent messagebar - pms@romance.kaist.ac.kr 
*/
static int ci_previous_undeleted (void)
{
  short i, j;

  i = (Context->current>0)? Context->current-1: Context->vcount-1;
  for(j=1; Context->hdrs[msgno(i)]->deleted; i--, j++) {
    if(i==0) i = Context->vcount;
    if(j==Context->vcount-1) { if(i==Context->vcount) i = -1; break; }
  }
  return(i);
}

static int ci_save_message (int delete, int ignore_selection)
{
  int redraw = 0;
  char ch, buffer[_POSIX_PATH_MAX];

  /* Do this in case we get called from process_command().  */
  move(LINES-3, 0);
  clrtobot();

  mutt_default_savename(buffer, Context->hdrs[*Context->v2r[Context->current]]);
  mvprintw(LINES-2, 0, "%s to folder (? for list): %s",
	   delete ? "Save" : "Copy", buffer);
  refresh();
  ch = ci_getch();
  if (ch == '?') {
    folder_select(Maildir, buffer);
    redraw = REDRAW_FULL;
  }
  else if (ch == ctrl('G'))
    return 0;
  else if (!CI_is_return(ch)) {
    if (isprint(ch)) {
      buffer[0] = ch;
      buffer[1] = 0;
      ch = 0;
    }
    ci_get_field(delete ? "Save to folder: " : "Copy to folder: ", buffer, ch,
		 CI_complete);
    if (!buffer[0]) {
      ClearLine(LINES-2);
      return 0;
    }
  }

  /*
   * This is an undocumented feature of ELM pointed out to me by Felix von
   * Leitner <leitner@prz.fu-berlin.de>
   */
  if (strcmp(buffer, ".") == 0)
    strcpy(buffer, ci_last_folder);
  else
    strcpy(ci_last_folder, buffer);

  if (!ignore_selection && Context->selected)
    mutt_save_message(0, buffer, delete);
  else
    mutt_save_message(Context->hdrs[*Context->v2r[Context->current]], buffer,
		      delete);

  ClearLine(LINES-2);
  return(redraw);
}

static void ci_bounce_message (void)
{
  char buf[STRING];
  ADDRESS *adr = 0, *padr;

  buf[0] = 0;
  ci_get_field("Bounce message to: ", buf, 0, 0);
  if (buf[0] == 0) return;
  rfc822_parse_adrlist(&adr, buf, "@");
  adr = mutt_expand_aliases(adr);

  strcpy(buf, "Bounce message to ");
  rfc822_write_address(buf, adr);
  strcat(buf, "?");
  if (!mutt_yesorno(buf, 1)) {
    CI_clearline(LINES-2);
    return;
  }

  buf[0] = 0;
  padr = adr;
  while (padr) {
    sprintf(buf+strlen(buf), "%s@%s ", padr->mailbox, padr->host);
    padr = padr->next;
  }
  mutt_free_address(&adr);

  if (Context->selected)
    mutt_bounce_message(NULL, buf);
  else
    mutt_bounce_message(Context->hdrs[*Context->v2r[Context->current]], buf);
  CI_clearline(LINES-2);
  mutt_error("Message%s bounced.", Context->selected ? "s" : "");
}


static int ci_pipe_message (void)
{
  char buffer[STRING];

  buffer[0] = 0;
  ci_get_field("Pipe to command: ", buffer, 0, 0);
  if (buffer[0] == 0) return 0;
  endwin();
  mutt_pipe_message(Context->hdrs[*Context->v2r[Context->current]], buffer);
  ci_any_key_to_continue(0);
  return 1;
}

static int ci_select_sort_method (void)
{
  int reverse = Sort & SORTREVERSE, ch;
  char buffer[SHORT_STRING];

  Sort &= ~SORTREVERSE;

  CenterLine("<SPACE> = next value, <TAB> = Reverse, <RETURN> = select", LINES-1);

  FOREVER {
    switch (Sort & SORTMASK) {
    case SORTDATE:
      strcpy(buffer, "Date-Sent");
      break;
    case SORTSUBJECT:
    strcpy(buffer, "Subject");
      break;
    case SORTSIZE:
      strcpy(buffer, "Size");
      break;
    case SORTFROM:
      strcpy(buffer, "From");
      break;
    case SORTORDER:
      strcpy(buffer, "Date-Received");
      break;
    }
    CI_clearline(HDR_CMD);
    addstr("Sort By: ");
    if (reverse) addstr("Reverse-");
    addstr(buffer);
    ch = ci_getch();
    refresh();
    if (ch == ' ') {
      if (++Sort > SORTMAX) Sort = 0;
    }
    else if (ch == '\t')
      reverse = reverse ? 0 : SORTREVERSE;
    else if (CI_is_return(ch))
      break;
    else
      beep();
  }
  Sort |= reverse;
  ClearLine(HDR_CMD);
  return(0);
}

static void ci_show_title (void)
{
  char buffer[STRING];

  mvaddstr (1,0, Mutt_version);
  strcpy (buffer, "Folder: ");
  if (option(OPTSHOWHOST))
    sprintf(buffer+strlen(buffer), "%s@", Hostname);
    
  if (Context && Context->folder) {
    if (strncmp(Context->folder->name, Maildir, strlen (Maildir)) == 0)
      sprintf(buffer+strlen(buffer), "=%s ", Context->folder->name + strlen(Maildir) + 1);
    else if (strncmp(Context->folder->name, Homedir, strlen(Homedir)) == 0)
      sprintf(buffer+strlen(buffer), "~/%s ", Context->folder->name + strlen(Homedir) + 1);
    else
      sprintf(buffer+strlen(buffer), "%s ", Context->folder->name);
    sprintf(buffer+strlen(buffer), "(%d messages, %d new)", Context->msgcount, Context->new);
  }
  else
    strcat (buffer, "(no folder)");
  mvaddstr (1, COLS - strlen(buffer) - 1, buffer);
}

static int ci_set_top (void)
{
  int top;
  
  if ((top = Context->current - HDR_LEN) < 0) top = 0;
  return top;
}

/*
** To solve cusre display problem - pms@romance.kaist.ac.kr
*/
void AddMenuStr(unsigned char *s)
{
	register int sl, i, kor=0;
    register int cols = option(OPTARROWCURSOR)? COLS-2 : COLS;
	sl = strlen(s);
	if(sl<cols-1) for(i=sl; sl<cols-1; i++) s[i] = 0x20;
	for(i=0; i<cols-1; i++) {
		if(s[i] & 0x80) kor++; 
		else kor = 0;
	}
	if((i-kor)%2==0) s[cols-2] = 0x20; 			
	s[cols] = s[cols-1] = 0;
	addstr(s);
}

/*
 * This function handles the message index window as well as commands returned
 * from the pager (MENU_PAGER).
 */
void ci_main (void)
{
  volatile int top = 0;
  volatile int oldcurrent = Context->current;
  volatile int redraw = REDRAW_FULL;
  volatile int op;
  int menu=MENU_MAIN, done=0, i=0;
  char buffer[STRING];

  ci_last_folder[0] = 0;

#ifdef NCURSES_MOUSE_VERSION
  if (Mousemask) toolbar_init(Main_toolbar);
#endif

  Errorbuf[0] = 0;
  top = ci_set_top();

  FOREVER {
    
      if (ci_sigwinch) {
   endwin();
   initscr();
   ci_sigwinch = 0;
   redraw = REDRAW_FULL;
   menu = MENU_MAIN;
      }

    if (menu == MENU_MAIN) {
      /* check for new mail in the mailbox */
      if (mutt_check_mailbox(Context)) redraw = REDRAW_FULL;

      if (redraw == REDRAW_FULL) {
	clear();
	ci_show_title();
#ifdef NCURSES_MOUSE_VERSION
	if (Mousemask) toolbar_post(Main_toolbar);
#endif
	ci_show_last_error();
	redraw = REDRAW_INDEX;
      }

      if (Context->folder) {
	if (Context->current > top+HDR_LEN) {
	  top = Context->current;
	  redraw = REDRAW_INDEX;
	}
	else if (Context->current < top) {
	  if ((top = Context->current - HDR_LEN) < 0) top = 0;
	  redraw = REDRAW_INDEX;
	}
	else if (oldcurrent != Context->current) {
	  if (!redraw) redraw = REDRAW_MOTION;
	}
      }

      if (redraw == REDRAW_INDEX) {
	for (i=top; i<=top+HDR_LEN; i++) {
	  if (i<Context->vcount) {
	    CI_generate_header(buffer, Context->hdrs[*Context->v2r[i]], sizeof(buffer)-1);
	    CI_clearline(i-top+HDR_OFFSET);
	    
	    if (option(OPTARROWCURSOR))
	      move(i-top+HDR_OFFSET, 2);
	    else if (i==Context->current)
	      attrset(A_REVERSE);

	    AddMenuStr(buffer);

	    if (!option(OPTARROWCURSOR) && i==Context->current)
	      attrset(A_NORMAL);
	  }
	  else {
	    CI_clearline(i-top+HDR_OFFSET);
	  }
	}
	redraw = 0;
      }
      else if (redraw == REDRAW_MOTION) {
	if (option(OPTARROWCURSOR)) {
	  CI_generate_header(buffer, Context->hdrs[oldcurrent], sizeof(buffer)-1);
	  mvaddstr(oldcurrent-top+HDR_OFFSET, 0, "  ");
	  AddMenuStr(buffer);
	}
	else {
	  /* redraw the current line in case the flags changed. */
	  CI_clearline(oldcurrent-top+HDR_OFFSET);
	  CI_generate_header(buffer, Context->hdrs[*Context->v2r[oldcurrent]], sizeof(buffer)-1);
	  AddMenuStr(buffer);

	  move(Context->current-top+HDR_OFFSET, 0);
	  CI_generate_header(buffer, Context->hdrs[*Context->v2r[Context->current]], sizeof(buffer)-1);
	  attrset(A_REVERSE);
	  AddMenuStr(buffer);
	  attrset(A_NORMAL);
	}

	redraw = 0;
      }
      else if (redraw == REDRAW_CURRENT) {	  
	if (option(OPTARROWCURSOR)) {
	  move(Context->current-top+HDR_OFFSET, 2);
	}
	else  {
	  move(Context->current-top+HDR_OFFSET, 0);
	  attrset(A_REVERSE);
	}

	CI_generate_header(buffer, Context->hdrs[*Context->v2r[Context->current]], sizeof(buffer)-1);
	AddMenuStr(buffer);

	if (!option(OPTARROWCURSOR)) { attrset(A_NORMAL); }

	redraw = 0;
      }
      oldcurrent = Context->current;

      if (option(OPTARROWCURSOR)) {
	move(Context->current-top+HDR_OFFSET, 0);
	if (Context->msgcount) {
	  attrset(A_REVERSE);
	  addstr("->");
	  attrset(A_NORMAL);
	}
      }
      else {
	move(Context->current-top+HDR_OFFSET, COLS-1);
      }

      refresh();

      if ((i = ci_getch()) == ERR) continue;

      ci_clear_error();

#ifdef NCURSES_MOUSE_VERSION
      if (i == KEY_MOUSE) {
	MEVENT e;

	getmouse(&e);
	if (e.y == 0) {
	  /* This click occurred on the toolbar. */
	  if (e.bstate & BUTTON1_CLICKED) {
	    if ((op = toolbar_select(Main_toolbar, e.x)) == -1) continue;
	  }
	}
	else if (e.bstate & BUTTON1_CLICKED) {
	  if (e.y >= HDR_OFFSET && e.y <= HDR_OFFSET+HDR_LEN) {
	    Context->current = e.y - HDR_OFFSET + top;
	    if (Context->current > Context->vcount - 1)
	      Context->current = oldcurrent;
	  }
	  continue;
	}
	else if (e.bstate & BUTTON1_DOUBLE_CLICKED) {
	  if (e.y >= HDR_OFFSET && e.y <= HDR_OFFSET+HDR_LEN) {
	    Context->current = e.y - HDR_OFFSET + top;
	    if (Context->current > Context->vcount - 1) {
	      Context->current = oldcurrent;
	      continue;
	    }
	    op = ci_display_message();
	    menu = MENU_PAGER;
	  }
	  continue;
	}
      }
      else
#endif /* NCURSES_MOUSE_VERSION */
	op = keymap_lookup(i, MENU_MAIN);
    }
    else { /* menu != MENU_MAIN */
      move(LINES-3, 0);
      clrtobot();
    }

    switch (op) {

    case OP_MAIN_QUIT:
      mutt_close_mailbox(Context);
      done = 1;
      break;

    case OP_MAIN_REFRESH:
      redraw = REDRAW_FULL;
      break;
    
    case OP_MAIN_HELP:
      if (help(MENU_MAIN)) redraw = REDRAW_FULL;
      break;
	
    case OP_MAIN_LAST_MESSAGE:
      if (Context->vcount) {
	Context->current = Context->vcount - 1;
	top = Context->current - HDR_LEN;
	if (top < 0)
	  top = 0;
	redraw = REDRAW_INDEX;
      }
      break;

    case OP_MAIN_FIRST_MESSAGE:
      Context->current = 0;
      break;
      
    case OP_MAIN_NEXT_PAGE:
      if (Context->vcount && (top + HDR_LEN + 1 < Context->vcount))
	Context->current = top + HDR_LEN + 1;
      else
	beep();
      break;

    case OP_MAIN_PREVIOUS_PAGE:
      if (Context->vcount && top > 0)
	Context->current = top - 1;
      else
	beep();
      break;

    case OP_MAIN_DELETE_PATTERN:
      if (Context->vcount) {
	pattern_function(ACTIONDELETE, "Delete messages matching: ");
	redraw = REDRAW_INDEX;
      }
      break;

    case OP_MAIN_UNDELETE_PATTERN:
      if (Context->vcount) {
	pattern_function(ACTIONUNDELETE, "Undelete messages matching: ");
	redraw = REDRAW_INDEX;
      }
      break;

    case OP_MAIN_TAG_MESSAGE:
      if (Context->vcount) {
	if (Context->hdrs[msgno(Context->current)]->selected) {
	  Context->hdrs[msgno(Context->current)]->selected = 0;
	  Context->selected--;
	}
	else {
	  Context->hdrs[msgno(Context->current)]->selected = 1;
	  Context->selected++;
	}
	if (Context->current < Context->vcount - 1)
	  Context->current++;
	else
	  redraw = REDRAW_CURRENT;
      }
      break;

    case OP_MAIN_TAG_PATTERN:
      if (Context->vcount) {
	pattern_function(ACTIONSELECT, "Tag messages matching: ");
	redraw = REDRAW_INDEX;
      }
      break;

    case OP_MAIN_JUMP:
      if (Context->vcount) {
	buffer[0] = 0;
	ci_get_field("Jump to message: ", buffer, i, 0);
	if (buffer[0]) {
	  int n = atoi(buffer);

	  if (n > 0 && n <= Context->msgcount) {
	    if (Context->hdrs[n-1]->virtual != -1)
	      Context->current = Context->hdrs[n-1]->virtual;
	    else
	      mutt_error("That message is not visible.");
	  }
	  else
	    mutt_error("Invalid message number.");
	}
      }
      break;

    case OP_MAIN_SEARCH:
      if (Context->vcount)
	ci_search_command();
      break;

    case OP_MAIN_LIMIT:
      if (Context->vcount) {
	pattern_function(ACTIONLIMIT, "Limit to messages matching: ");
	Context->current = 0;
	redraw = REDRAW_INDEX;
      }
      break;

    case OP_MAIN_VIEW_ATTACHMENTS:
      if (Context->vcount) {
	if (Context->hdrs[msgno(Context->current)]->content->type == TYPEMULTIPART) {
	  mutt_parse_mime_message(Context->hdrs[*Context->v2r[Context->current]]);	
	  ci_attach();
	  redraw = REDRAW_FULL;
	}
	else
	  mutt_error("This is not a multipart message.");
      }
      break;

    case OP_MAIN_SORT_MESSAGES:
      ci_select_sort_method();
      if (Context->vcount) {
	mutt_sort_headers(Context);
	redraw = REDRAW_INDEX;
      }
      break;

    case OP_MAIN_SYNC_FOLDER:
      if (Context->vcount) {
	strcpy(buffer, Context->folder->name);
	if (mutt_sync_mailbox(Context)) {
	  Context = mutt_open_mailbox(buffer, Context);
	  top = ci_set_top();
	}
	redraw = REDRAW_FULL;
      }
      break;

    case OP_MAIN_UNDELETE_MESSAGE:
      if (Context->vcount) {
	Context->hdrs[msgno(Context->current)]->deleted = 0;
	if (Context->current < Context->vcount-1)
	  Context->current++;
	else
	  redraw = REDRAW_CURRENT;
      }
      break;

    /***********************************************************************
     * The following operations can also be performed inside of the pager. *
     ***********************************************************************/

    case OP_MAIN_EXIT:
      if (menu == MENU_MAIN) {
	mutt_fastclose_mailbox(Context);
	done = 1;
      }
      break;

    case OP_MAIN_DISPLAY_MESSAGE:
    case OP_MAIN_DISPLAY_HEADERS: /* don't weed the headers */
      if (Context->vcount) {
	if (op == OP_MAIN_DISPLAY_HEADERS)
	  Context->weed = 0;
	else
	  Context->weed = 1;
	op = ci_display_message();
	menu = MENU_PAGER;
	continue;
      }
      break;

    case OP_MAIN_NEXT_UNDELETED:
      if (Context->vcount) {
	if (Context->current < Context->vcount - 1) {
	  if ((Context->current = ci_next_undeleted()) == -1) {
	    Context->current = oldcurrent;
	    mutt_error("No undeleted messages.");
	  }
	  else if (menu == MENU_PAGER) {
	    op = ci_display_message();
	    continue;
	  }
	}
	else if (menu == MENU_MAIN)
      Context->current = 0;
/* PMS
	  mutt_error("You are on the last message.");
*/
      }
      break;

    case OP_MAIN_NEXT_MESSAGE:
      if (Context->vcount) {
	if (Context->current < Context->vcount - 1) {
	  Context->current++;
	  if (menu == MENU_PAGER) {
	    op = ci_display_message();
	    continue;
	  }
	}
	else if (menu == MENU_MAIN)
/* PMS
	  mutt_error("You are on the last message.");
*/
      Context->current = 0;
      }
      break;

    case OP_MAIN_PREVIOUS_UNDELETED:
      if (Context->vcount) {
	if (Context->current > 0) {
	  if ((Context->current = ci_previous_undeleted()) == -1) {
	    Context->current = oldcurrent;
	    mutt_error("No undeleted messages.");
	  }
	  else if (menu == MENU_PAGER) {
	    op = ci_display_message();
	    continue;
	  }
	}
	else if (menu == MENU_MAIN)
/* PMS
	  mutt_error("You are on the first message.");
*/
      Context->current = Context->vcount-1;
      }
      break;

    case OP_MAIN_PREVIOUS_MESSAGE:
      if (Context->vcount) {
	if (Context->current > 0) {
	  Context->current--;
	  if (menu == MENU_PAGER) {
	    op = ci_display_message();
	    continue;
	  }
	}
	else if (menu == MENU_MAIN)
/* PMS - BOTTOM
      mutt_error("You are on the first message.");
*/
      Context->current = Context->vcount-1;
      }
      break;

    case OP_MAIN_ALIAS:
      if (Context->vcount)
	mkalias();
      break;

    case OP_MAIN_BOUNCE_MESSAGE:
      if (Context->vcount)
	ci_bounce_message();
      break;

    case OP_MAIN_CHANGE_FOLDER:
      mvaddstr(HDR_CMD, 0, "Open folder ('?' for list): ");
      refresh();
      i = ci_getch();
      CI_clearline(HDR_CMD);
      if (CI_is_return(i) || i == ctrl('G'))
	break;
      if (i != '?') {
	buffer[0] = 0;
	ci_get_field("Open folder: ", buffer, i, CI_complete);
      }
      else {
	folder_select(Maildir, buffer);
	redraw = REDRAW_FULL;
      }
      if (buffer[0] == 0) break;
      mutt_close_mailbox(Context);
      if ((Context = mutt_open_mailbox(buffer, Context)) == NULL)
	mutt_error("Unable to open %s!", buffer);
      top = ci_set_top();
      redraw = REDRAW_FULL;
      break;

    case OP_MAIN_COPY_MESSAGE:
    case OP_MAIN_SAVE_MESSAGE:
      if (Context->vcount) {
	redraw = ci_save_message((op == OP_MAIN_SAVE_MESSAGE), 0);
    if (!redraw && op == OP_MAIN_SAVE_MESSAGE && menu == MENU_MAIN) {
	  /*
	   * Must redraw the screen in some form since messages will
	   * have been deleted by save_to_folder().
	   */
	  if (!Context->selected) {
	    if ((Context->current = ci_next_undeleted()) == -1) {
	      Context->current = oldcurrent;
	      redraw = REDRAW_CURRENT;
	    }
	  }
	  else
	    redraw = REDRAW_INDEX;
	}
      }
      break;

    case OP_MAIN_DELETE_MESSAGE:
      if (Context->vcount) {
	Context->hdrs[msgno(Context->current)]->deleted = 1;
	if ((Context->current = ci_next_undeleted()) == -1) {
	  Context->current = oldcurrent;

	  if (menu == MENU_MAIN)
	    redraw = REDRAW_CURRENT;
	  else {
	    menu = MENU_MAIN;
	    redraw = REDRAW_FULL;
	  }
	}
	else if (menu == MENU_PAGER) {
	  oldcurrent = Context->current;
	  op = ci_display_message();
	  continue;
	}
      }
      break;

    case OP_MAIN_FORWARD_MESSAGE:
      if (Context->vcount) {
	ci_send_message(0, 0, SENDFORWARD);
	redraw = REDRAW_FULL;
      }
      break;

    case OP_MAIN_REPLY:
      if (Context->vcount) {
	ci_send_message(0, 0, SENDREPLY);
	redraw = REDRAW_FULL;
      }
      break;

    case OP_MAIN_GROUP_REPLY:
      if (Context->vcount) {
	ci_send_message(0, 0, SENDREPLY|SENDGROUPREPLY);
	redraw = REDRAW_FULL;
      }
      break;

    case OP_MAIN_LIST_REPLY:
      if (Context->vcount) {
	ci_send_message(0, 0, SENDREPLY|SENDLISTREPLY);
	redraw = REDRAW_FULL;
      }
      break;

    case OP_MAIN_MAIL:
      ci_send_message(0, 0, 0);
      redraw = REDRAW_FULL;
      break;
    
    case OP_MAIN_PIPE_MESSAGE:
      if (Context->vcount) {
	if (ci_pipe_message())
	  redraw = REDRAW_FULL;
      }
      break;

    case OP_MAIN_PRINT_MESSAGE:
      if (Context->vcount)
	mutt_print_message(Context->hdrs[*Context->v2r[Context->current]]);
      break;

    case OP_MAIN_FLAG_MESSAGE:
#define TOGGLE_FLAG(x) (x)->flagged = !(x)->flagged; (x)->changed = 1

      if (Context->vcount) {
	TOGGLE_FLAG(Context->hdrs[*Context->v2r[Context->current]]);
	if (menu == MENU_MAIN) redraw = REDRAW_CURRENT;
      }
      break;

    case OP_MAIN_MARK_AS_NEW:
      if (Context->vcount) {
	Context->hdrs[*Context->v2r[Context->current]]->read = 0;
	Context->hdrs[*Context->v2r[Context->current]]->old = 0;
	Context->hdrs[*Context->v2r[Context->current]]->changed = 1;
	if (menu == MENU_MAIN) redraw = REDRAW_CURRENT;
      }
      break;

#ifdef _PGPPATH
    case OP_MAIN_FORGET_PASSPHRASE:
      pgp_void_passphrase();
      mutt_error("PGP passphrase forgotten.");
      break;
#endif /* _PGPPATH */

    default:
      if (menu == MENU_MAIN) {
   mutt_error("Key is not bound.  Press '?' for help.");
    beep();
      }
    }

    if (menu == MENU_PAGER) {
      menu = MENU_MAIN;
      if (!redraw) redraw = REDRAW_FULL;
      Context->weed = 1; /* turn header weeding back on. */
    }

    if (done) break;
  }

  move(LINES-1, 0);
  clrtoeol();
  endwin();

  printf("HanMutt has completed.\n");
}
