/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "muttlib.h"
#include "rfc822.h"
#include "rfc1522.h"

int mutt_is_mail_list (ADDRESS *addr)
{
  LIST *p = Mailing_lists;

  if (!addr->mailbox) return 0;
  while (p) {
    if (strncasecmp(addr->mailbox, p->data, strlen(p->data)) == 0)
      return 1;
    p = p->next;
  }
  return 0;
}

static void make_from (char *buffer, HEADER *hdr, unsigned int len)
{
  ADDRESS *tmpaddr;

  buffer[0] = 0;
  if (mutt_addr_is_user(hdr->from)) {
    strcpy(buffer, "To ");
    rfc822_write_address(buffer + sizeof("To ") - 1, hdr->to);
    return;
  }

  tmpaddr = hdr->to;
  while (tmpaddr) {
    if (mutt_is_mail_list(tmpaddr)) {
      sprintf(buffer, "To %s", tmpaddr->mailbox);
      return;
    }
    tmpaddr = tmpaddr->next;
  }

  tmpaddr = hdr->cc;
  while (tmpaddr) {
    if (mutt_is_mail_list(tmpaddr)) {
      sprintf (buffer, "Cc %s", tmpaddr->mailbox);
      return;
    }
    tmpaddr = tmpaddr->next;
  }

  if (hdr->from && hdr->from->personal) {
    strfcpy(buffer, hdr->from->personal, len);
  }
  else {
    if (option(OPTREVALIAS)) {
      ADDRESS *t = alias_reverse_lookup(hdr->from);

      if (t && t->personal) {
        strcpy(buffer, t->personal);
        return;
      }
    }
      
    buffer[0] = 0; /* otherwise the string will be appended... */
    rfc822_write_address(buffer, hdr->from);
  }
}

static int user_is_recipient (HEADER *hdr)
{
  ADDRESS *a;

  a = hdr->to;
  while (a) {
    if (mutt_addr_is_user(a)) return(1);
    a=a->next;
  }
  a = hdr->cc;
  while (a) {
    if (mutt_addr_is_user(a)) return(1);
    a=a->next;
  }
  return(0);
}

/*
 * %a = address of author
 * %c = size of message in bytes
 * %d = date and time of message
 * %f = entire from line
 * %l = number of lines in the message (MUST have: unset honor_length)
 * %L = list-from function.
 * %n = name of author
 * %i = message-id
 * %u = user (login) name of author
 * %M = month name of message
 * %m = number of month
 * %N = number of the day of the month
 * %s = subject
 * %Y = year
 * %y = last two digits of the year
 *
 */

/*
**  - modified by pms@romance.kaist.ac.kr
*/
void mutt_make_string (char *dest, char *s, HEADER *hdr, int destlen)
{
  extern void DecodeHeader(char*, const char*);
  char buf[512], buf2[512], prefix[16], *pprefix, fmt[32], ts[512];
  int len;

  destlen--; /* save room for the terminal null (\0) character */
  while (*s && destlen>0) {
    if (*s == '%') {
      s++;
      
      if (*s == '%') {
	*dest++ = '%';
	destlen--;
	continue;
      }

      /* strip off the formatting commands */
      pprefix=prefix;
      while (*s == '-' || *s == '.' || isdigit(*s)) {
	*pprefix++ = *s++;
      }
      *pprefix = 0;
      switch(*s) {
      case 'a':
	sprintf(buf2, "%s@%s", hdr->from->mailbox, hdr->from->host);
	sprintf(fmt, "%%%ss", prefix);
	sprintf(buf, fmt, buf2);
	break;
      case 'f':
	buf2[0] = 0;
/* PMS */
    if(hdr->from) rfc822_write_address(buf2, hdr->from);
    else if(hdr->to) {
        strcpy(buf2, "To ");
        rfc822_write_address(buf2, hdr->to);
    }
    sprintf(fmt, "%%%ss", prefix);
    DecodeHeader(ts, buf2);
	sprintf(buf, fmt, buf2);
	break;
      case 'i':
   sprintf(fmt, "%%%ss", prefix);
        sprintf(buf, fmt, hdr->message_id ? hdr->message_id : "<no.id>");
   break;
      case 'l':
   sprintf(fmt, "%%%sd", prefix);
   sprintf(buf, fmt, hdr->lines);
   break;
      case 'L':
	make_from(buf2, hdr, sizeof(buf2));
	sprintf(fmt, "%%%ss", prefix);
	sprintf(buf, fmt, buf2);
	break;
      case 'n':
	if (hdr->from && hdr->from->personal)
	  strcpy(buf2, hdr->from->personal);
/* PMS */
    else if(hdr->from) {
      buf2[0] = 0;
      rfc822_write_address(buf2, hdr->from);
    }
    else if(hdr->to) {
        if(hdr->to->personal) sprintf(buf2, "To %s", hdr->to->personal);
        else {
            strcpy(buf2, "To ");
            rfc822_write_address(buf2, hdr->to);
        }
    }
    else buf2[0] = 0;
    sprintf(fmt, "%%%ss", prefix);
    DecodeHeader(ts, buf2);
    sprintf(buf, fmt, ts);
	break;
      case 'u':
	sprintf(fmt, "%%%ss", prefix);
	sprintf(buf, fmt, hdr->from->mailbox);	
	break;
      case 'd':
	sprintf(buf2, "%s %d, %d %d:%02d:%02d %c%02d%02d",
		Months[hdr->month], hdr->day, hdr->year+1971, hdr->hours,
		hdr->minutes, hdr->seconds, hdr->zoccident ? '-' : '+',
		hdr->zhours, hdr->zminutes);
	sprintf(fmt, "%%%ss", prefix);
	sprintf(buf, fmt, buf2);
	break;
      case 'M':
	sprintf(fmt, "%%%ss", prefix);
	sprintf(buf, fmt, Months[hdr->month]);
	break;
      case 's':
	sprintf(fmt, "%%%ss", prefix);
	sprintf(buf, fmt, hdr->subject ? hdr->subject : "< no subject >");
	break;
      case 'c':
	sprintf(fmt, "%%%sd", prefix);
	sprintf(buf, fmt, hdr->length);
	break;
      case 'm':
	sprintf(fmt, "%%%sd", prefix);
	sprintf(buf, fmt, hdr->month+1);
	break;
      case 'N':
	sprintf(fmt, "%%%sd", prefix);
	sprintf(buf, fmt, hdr->day);
	break;
      case 'Y':
	sprintf(fmt, "%%%sd", prefix);
	sprintf(buf, fmt, hdr->year + 1971);
	break;
      case 'y':
	sprintf(fmt, "%%%sd", prefix);
	sprintf(buf, fmt, (hdr->year + 1971) % 100);
	break;
      }

      if ((len = strlen(buf)) < destlen) {
	strcpy(dest, buf);
	dest += len;
	destlen -= len;
      }
      else {
	strfcpy(dest, buf, destlen);
	return; /* no more space! */
      }
    }
    else if (*s == '\\') {
      s++;
      if (*s == 'n') {
	*dest++ = '\n';
	destlen--;
      }
      else if (*s == 't') {
	*dest++ = '\t';
	destlen--;
      }
      else {
	*dest++ = '?';
	destlen--;
      }
    }
    else {
      *dest++ = *s;
      destlen--;
    }
    s++;
  }
  *dest = 0;
}

/*
**  Inform multipart mail - modified by pms@romance.kaist.ac.kr
*/
void mutt_generate_header (char *b, HEADER *cur, int blen)
{
  char from[LONG_STRING]; /* has to be long enough to fit the To: line... */
  char flag;

  make_from(from, cur, sizeof(from));

  if (cur->metamail)
    flag = 'M';
#ifdef _PGPPATH
  else if (cur->pgp == PGPSIGN)
    flag = 'S';
  else if (cur->pgp == PGPENCRYPT)
    flag = 'P';
  else if (cur->pgp == PGPKEY)
    flag = 'K';
#endif /* _PGPPATH */
  else if(cur->content->type == TYPEMULTIPART)
    flag = 'v';
  else
    flag = ' ' ;

  /* first do the fixed part of the header */
  sprintf(b, "%4d%c%c%c%c",
	  cur->msgno,
	  cur->read ? (cur->replied ? 'r' : ' ') : (cur->old ? 'O' : 'N'),
	  cur->deleted ? 'D' : ' ',
	  flag,
	  cur->selected ? 'X' : (cur->flagged ? '!' : (user_is_recipient(cur) ? '+' : ' '))
	  );

  /* now process the user configurable part. */
  mutt_make_string(b+strlen(b), Hdr_format, cur, blen);
}
