/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "keymap.h"

static char *help_lookupFunction (int op, int menu)
{
  int i;

  switch (menu) {
  case MENU_MAIN:

    for (i=0; Op_main[i].name; i++)
      if (Op_main[i].op == op)
        return(Op_main[i].name);
    break;

  case MENU_SEND:

    for (i=0; Op_send[i].name; i++)
      if (Op_send[i].op == op)
        return(Op_send[i].name);
    break;

  default:

    for (i=0; Op_generic[i].name; i++) {
      if (Op_generic[i].op == op)
        return(Op_generic[i].name);    
    }
    if (menu == MENU_ATTACH) {
      for (i=0; Op_attach[i].name; i++)
        if (Op_attach[i].op == op)
  	return(Op_attach[i].name);
    }
  }
  
  return 0;

}
 
static char *help_keyname (int c)
{
  static char buf[2];

  if (c == ' ')
    return "<space>";
  else if (c == '\n')
    return "<return>";
  else if (c == '\t')
    return "<tab>";
  else {
    sprintf(buf,"%c",(unsigned char)c);
    return buf;
  }
}

static void help_listKeys (int menu)
{
  char *t = mutt_new_tempfile();
  char *desc;
  struct keymap *k;
  FILE *f;
  int i;

  f = fopen(t, "w");
  for (i = 0; i < 128; i++) {
    if ((k = Keymap[i])) {
      while (k) {
        if (k->menu == menu) {
	  desc = help_lookupFunction(k->function, menu);
          fprintf(f, "\t%s\t%s\n", help_keyname(i),
         desc ? desc : "<ERROR!  Missing definition for binding!>");
          break;
        }
        k = k->next;
      }
    }
  }
  fclose(f);
  ci_simple_pager(t);
  safe_free((void **)&t);
}

/* returns 1 if screen needs to be redrawn */
int help (int menu)
{
  int ch;

  mvaddstr(LINES-2, 0, "Enter key ('?' = list all bindings): ");
  refresh();
  ch = getch();
  if (ch == '?') {
    help_listKeys(menu);
    return 1;
  }
  else {
    KEYMAP *k;

    if (ch < MAX_KEYS)
      k = Keymap[ch];
    else
      k = 0;
    move(LINES-2,0);
    clrtoeol();
    while (k) {
      if (k->menu == menu) {
        mutt_error("%s is bound to function \"%s\".", help_keyname(ch),
		help_lookupFunction(k->function, menu));
        break;
      }
      k = k->next;
    }
    if (!k) {
      mutt_error("Key is not bound to a function.");
    }
    return 0;
  }
}
