/* Copyright (C) 1996 Thomas 'Mike' Michlmayr
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Modified by Michael Elkins <me@cs.hmc.edu>
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "keymap.h"

/* global hashtable for keys (menus are in linked lists starting here) */
KEYMAP *Keymap[MAX_KEYS];

/*
 * internal function used to insert a binding into the linked list
 */
void keymap_bind (char ch, int menu, int function) {
  KEYMAP *p = Keymap[(int)ch];
  
  if (!p) {
    p = Keymap[(int)ch] = (KEYMAP *)safe_malloc(sizeof(KEYMAP));
    p->menu = menu;
    p->function = function;
    p->next = 0;
    return;
  }

  /* need to get rid of any previous bindings for this key in this menu */
  
  FOREVER {
    if (p->menu == menu) {
      /* this binding overrides the previous binding. */
      p->function = function;
      return;
    }
    if (!p->next) {
      /* no binding exists, so create a new one. */
      p->next = (KEYMAP *)safe_malloc(sizeof(KEYMAP));
      p = p->next;
      p->menu = menu;
      p->function = function;
      p->next = 0;
      return;
    }
    p = p->next;
  }

  /* not reached */
}

/*
 * This function generates the default keybindings for a given menu.
 */

static void generic_bindings (int menu) {
  keymap_bind('?', menu, OP_GENERIC_HELP);
  keymap_bind('\n', menu, OP_GENERIC_SELECT_ENTRY);
  keymap_bind('j', menu, OP_GENERIC_NEXT_ENTRY);
  keymap_bind('k', menu, OP_GENERIC_PREVIOUS_ENTRY);
  keymap_bind('h', menu, OP_GENERIC_PREVIOUS_PAGE);
  keymap_bind('l', menu, OP_GENERIC_NEXT_PAGE);
  keymap_bind('q', menu, OP_GENERIC_EXIT);
  keymap_bind('x', menu, OP_GENERIC_EXIT);
  keymap_bind('=', menu, OP_GENERIC_FIRST_ENTRY);
  keymap_bind('*', menu, OP_GENERIC_LAST_ENTRY);
  keymap_bind(ctrl('L'), menu, OP_GENERIC_REFRESH);
}

/* 
 * for now we just define this "by hand". if we have a parser, we expand
 * this function.
 */

void keymap_init (void) {
  int i;

  for (i=0;i<MAX_KEYS;i++) {
    Keymap[i] = NULL;
  }
  (void) keymap_bind('q', MENU_MAIN, OP_MAIN_QUIT);
  (void) keymap_bind('Q', MENU_MAIN, OP_MAIN_QUIT);
  (void) keymap_bind('x', MENU_MAIN, OP_MAIN_EXIT);
  (void) keymap_bind('m', MENU_MAIN, OP_MAIN_MAIL);
  (void) keymap_bind('c', MENU_MAIN, OP_MAIN_CHANGE_FOLDER);
  (void) keymap_bind(ctrl('L'), MENU_MAIN, OP_MAIN_REFRESH);
  (void) keymap_bind('a', MENU_MAIN, OP_MAIN_ALIAS);
  (void) keymap_bind('b', MENU_MAIN, OP_MAIN_BOUNCE_MESSAGE);
  (void) keymap_bind('j', MENU_MAIN, OP_MAIN_NEXT_UNDELETED);
  (void) keymap_bind('n', MENU_MAIN, OP_MAIN_NEXT_UNDELETED);
  (void) keymap_bind('J', MENU_MAIN, OP_MAIN_NEXT_MESSAGE);
  (void) keymap_bind('N', MENU_MAIN, OP_MAIN_NEXT_MESSAGE);
  (void) keymap_bind('k', MENU_MAIN, OP_MAIN_PREVIOUS_UNDELETED);
  (void) keymap_bind('K', MENU_MAIN, OP_MAIN_PREVIOUS_MESSAGE);
  (void) keymap_bind('*', MENU_MAIN, OP_MAIN_LAST_MESSAGE);
  (void) keymap_bind('=', MENU_MAIN, OP_MAIN_FIRST_MESSAGE);
  (void) keymap_bind('z', MENU_MAIN, OP_MAIN_NEXT_PAGE);
  (void) keymap_bind('Z', MENU_MAIN, OP_MAIN_PREVIOUS_PAGE);
  (void) keymap_bind(' ', MENU_MAIN, OP_MAIN_DISPLAY_MESSAGE);
  (void) keymap_bind('\n', MENU_MAIN, OP_MAIN_DISPLAY_MESSAGE);
  (void) keymap_bind('h', MENU_MAIN, OP_MAIN_DISPLAY_HEADERS);
  (void) keymap_bind('d', MENU_MAIN, OP_MAIN_DELETE_MESSAGE);
  keymap_bind('D', MENU_MAIN, OP_MAIN_DELETE_PATTERN);
  (void) keymap_bind('u', MENU_MAIN, OP_MAIN_UNDELETE_MESSAGE);
  keymap_bind('U', MENU_MAIN, OP_MAIN_UNDELETE_PATTERN);
  (void) keymap_bind('$', MENU_MAIN, OP_MAIN_SYNC_FOLDER);
  (void) keymap_bind('r', MENU_MAIN, OP_MAIN_REPLY);
  (void) keymap_bind('g', MENU_MAIN, OP_MAIN_GROUP_REPLY);
  (void) keymap_bind('t', MENU_MAIN, OP_MAIN_TAG_MESSAGE);
  (void) keymap_bind('T', MENU_MAIN, OP_MAIN_TAG_PATTERN);
  (void) keymap_bind('/', MENU_MAIN, OP_MAIN_SEARCH);
  (void) keymap_bind('l', MENU_MAIN, OP_MAIN_LIMIT);
  (void) keymap_bind('p', MENU_MAIN, OP_MAIN_PRINT_MESSAGE);
  (void) keymap_bind('s', MENU_MAIN, OP_MAIN_SAVE_MESSAGE);
  (void) keymap_bind('C', MENU_MAIN, OP_MAIN_COPY_MESSAGE);
  (void) keymap_bind('f', MENU_MAIN, OP_MAIN_FORWARD_MESSAGE);
  (void) keymap_bind('o', MENU_MAIN, OP_MAIN_SORT_MESSAGES);
  (void) keymap_bind('|', MENU_MAIN, OP_MAIN_PIPE_MESSAGE);
  (void) keymap_bind('?', MENU_MAIN, OP_MAIN_HELP);
  (void) keymap_bind('v', MENU_MAIN, OP_MAIN_VIEW_ATTACHMENTS);
  keymap_bind('L', MENU_MAIN, OP_MAIN_LIST_REPLY);
  keymap_bind('F', MENU_MAIN, OP_MAIN_FLAG_MESSAGE);
  keymap_bind('N', MENU_MAIN, OP_MAIN_MARK_AS_NEW);
#ifdef _PGPPATH
  keymap_bind(ctrl('F'), MENU_MAIN, OP_MAIN_FORGET_PASSPHRASE);
#endif /* _PGPPATH */
  
  keymap_bind('1', MENU_MAIN, OP_MAIN_JUMP);
  keymap_bind('2', MENU_MAIN, OP_MAIN_JUMP);
  keymap_bind('3', MENU_MAIN, OP_MAIN_JUMP);
  keymap_bind('4', MENU_MAIN, OP_MAIN_JUMP);
  keymap_bind('5', MENU_MAIN, OP_MAIN_JUMP);
  keymap_bind('6', MENU_MAIN, OP_MAIN_JUMP);
  keymap_bind('7', MENU_MAIN, OP_MAIN_JUMP);
  keymap_bind('8', MENU_MAIN, OP_MAIN_JUMP);
  keymap_bind('9', MENU_MAIN, OP_MAIN_JUMP);
  
  /* defaults for the internal pager. */
  
  keymap_bind(' ',  MENU_PAGER, OP_PAGER_NEXT_PAGE);
  keymap_bind('-',  MENU_PAGER, OP_PAGER_PREVIOUS_PAGE);
  keymap_bind('\n', MENU_PAGER, OP_PAGER_NEXT_LINE);
  keymap_bind('^',  MENU_PAGER, OP_PAGER_TOP);
  keymap_bind('/',  MENU_PAGER, OP_PAGER_SEARCH);
  
  keymap_bind('j', MENU_PAGER, OP_MAIN_NEXT_UNDELETED);
  keymap_bind('J', MENU_PAGER, OP_MAIN_NEXT_MESSAGE);
  keymap_bind('n', MENU_PAGER, OP_MAIN_NEXT_UNDELETED);
  keymap_bind('k', MENU_PAGER, OP_MAIN_PREVIOUS_UNDELETED);
  keymap_bind('K', MENU_PAGER, OP_MAIN_PREVIOUS_MESSAGE);
  keymap_bind('b', MENU_PAGER, OP_MAIN_BOUNCE_MESSAGE);
  keymap_bind('m', MENU_PAGER, OP_MAIN_MAIL);
  keymap_bind('r', MENU_PAGER, OP_MAIN_REPLY);
  keymap_bind('g', MENU_PAGER, OP_MAIN_GROUP_REPLY);
  keymap_bind('d', MENU_PAGER, OP_MAIN_DELETE_MESSAGE);
  keymap_bind('u', MENU_PAGER, OP_MAIN_UNDELETE_MESSAGE);
  keymap_bind('q', MENU_PAGER, OP_MAIN_EXIT);
  keymap_bind('i', MENU_PAGER, OP_MAIN_EXIT);
  keymap_bind('x', MENU_PAGER, OP_MAIN_EXIT);
  keymap_bind('h', MENU_PAGER, OP_MAIN_DISPLAY_HEADERS);
  keymap_bind('C', MENU_PAGER, OP_MAIN_COPY_MESSAGE);
  keymap_bind('s', MENU_PAGER, OP_MAIN_SAVE_MESSAGE);
  keymap_bind('L', MENU_PAGER, OP_MAIN_LIST_REPLY);
  keymap_bind('|', MENU_PAGER, OP_MAIN_PIPE_MESSAGE);
  keymap_bind('f', MENU_PAGER, OP_MAIN_FORWARD_MESSAGE);
  keymap_bind('p', MENU_PAGER, OP_MAIN_PRINT_MESSAGE);
  keymap_bind('F', MENU_PAGER, OP_MAIN_FLAG_MESSAGE);

  /* defaults for the attachment menu. */

  generic_bindings(MENU_ATTACH);
  keymap_bind('s', MENU_ATTACH, OP_ATTACH_SAVE_ENTRY);
  keymap_bind('p', MENU_ATTACH, OP_ATTACH_PRINT_ENTRY);

  /* defaults for the send menu */

  keymap_bind('?', MENU_SEND, OP_SEND_HELP_MENU);
  keymap_bind('a', MENU_SEND, OP_SEND_ATTACH_FILE);
  keymap_bind('e', MENU_SEND, OP_SEND_EDIT_MESSAGE);
  keymap_bind('E', MENU_SEND, OP_SEND_EDIT_HEADERS);
  keymap_bind('v', MENU_SEND, OP_SEND_ALIAS_MENU);
  keymap_bind('c', MENU_SEND, OP_SEND_EDIT_CC);
  keymap_bind('b', MENU_SEND, OP_SEND_EDIT_BCC);
  keymap_bind('t', MENU_SEND, OP_SEND_EDIT_TO);
  keymap_bind('s', MENU_SEND, OP_SEND_EDIT_SUBJECT);
  keymap_bind('\n', MENU_SEND, OP_SEND_SEND_MESSAGE);
  keymap_bind('y', MENU_SEND, OP_SEND_SEND_MESSAGE);
  keymap_bind('n', MENU_SEND, OP_SEND_ABORT_MESSAGE);
  keymap_bind('P', MENU_SEND, OP_SEND_POSTPONE_MESSAGE);
#ifdef _PGPPATH
  keymap_bind('p', MENU_SEND, OP_SEND_PGP_MENU);
#endif
#ifdef MIXMASTER_PATH
  keymap_bind('m', MENU_SEND, OP_SEND_MIXMASTER_MENU);
#endif
#ifdef ISPELL
  keymap_bind('i', MENU_SEND, OP_SEND_ISPELL);
#endif

  generic_bindings(MENU_POST);
  keymap_bind('d', MENU_POST, OP_GENERIC_DELETE_ENTRY);
  keymap_bind('u', MENU_POST, OP_GENERIC_UNDELETE_ENTRY);

  /*
   * These menus only use the generic keybindings.
   */
  generic_bindings(MENU_FOLDER);
  generic_bindings(MENU_ALIAS);
  generic_bindings(MENU_REMAILER);
}

/*
 * most important, get function if you have the character and the menu
 */
int keymap_lookup (int ch, int menu) {
  
  if (ch < KEY_MIN) {
    KEYMAP *p = Keymap[ch];

    while (p) {
      if (p->menu == menu) return(p->function);
      p = p->next;
    }
    return(0);
  }

  /* 
   * Special handling for "curses" keys.  This may eventually be moved so
   * that each menu should check for these keys itself.
   */

  switch (menu) {

  case MENU_MAIN:
    switch (ch) {
    case KEY_HOME:
      return OP_MAIN_FIRST_MESSAGE;
    case KEY_END:
      return OP_MAIN_LAST_MESSAGE;
    case KEY_PPAGE:
    case KEY_LEFT:
      return OP_MAIN_PREVIOUS_PAGE;
    case KEY_NPAGE:
    case KEY_RIGHT:
      return OP_MAIN_NEXT_PAGE;
    case KEY_UP:
      return OP_MAIN_PREVIOUS_UNDELETED;
    case KEY_DOWN:
      return OP_MAIN_NEXT_UNDELETED;
    }
    break;

  case MENU_PAGER:
    switch (ch) {
    case KEY_HOME:
      return OP_PAGER_TOP;
    case KEY_PPAGE:
      return OP_PAGER_PREVIOUS_PAGE;
    case KEY_NPAGE:
      return OP_PAGER_NEXT_PAGE;
    case KEY_UP:
      return OP_MAIN_PREVIOUS_UNDELETED;
    case KEY_DOWN:
      return OP_MAIN_NEXT_UNDELETED;
    case KEY_LEFT:
      return OP_MAIN_PREVIOUS_UNDELETED;
    case KEY_RIGHT:
      return OP_MAIN_NEXT_UNDELETED;
    }
    break;
  }
  return 0;
}


/*
 * clean up and dealloc mem before leaving mutt
 */
void keymap_done(void) {
  int i;
  KEYMAP *p;
  KEYMAP *q;
  
  for (i=0;i<MAX_KEYS;i++) {
    p = Keymap[i];
    while (p != NULL) {
      q = p;
      p = p->next;
      safe_free((void **)&q);
    }
    Keymap[i] = NULL;
  }
}
