/* Copyright (C) 1996 Michael R. Elkins.
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define KEYMAP_GLOBAL_C

#include "keymap.h"

/*
 * Generic operations common to all menus.
 */
struct binding_t Op_generic[] = {
  { "select-entry",	OP_GENERIC_SELECT_ENTRY },
  { "next-entry",	OP_GENERIC_NEXT_ENTRY },
  { "previous-entry",	OP_GENERIC_PREVIOUS_ENTRY },
  { "last-entry",	OP_GENERIC_LAST_ENTRY },
  { "first-entry",	OP_GENERIC_FIRST_ENTRY },
  { "next-page",	OP_GENERIC_NEXT_PAGE },
  { "previous-page",	OP_GENERIC_PREVIOUS_PAGE },
  { "help",		OP_GENERIC_HELP },
  { "refresh",		OP_GENERIC_REFRESH },
  { "exit",             OP_GENERIC_EXIT },
  { "delete-entry",     OP_GENERIC_DELETE_ENTRY },
  { "undelete-entry",   OP_GENERIC_UNDELETE_ENTRY },
  { 0, 0 }
};

/*
 * Attachment Menu
 */
struct binding_t Op_attach[] = {
  { "display-entry", OP_GENERIC_SELECT_ENTRY },
  { "print-entry", OP_ATTACH_PRINT_ENTRY },
  { "save-entry", OP_ATTACH_SAVE_ENTRY },
  { "view-entry", OP_ATTACH_VIEW_ENTRY },
  { 0, 0 }
};

/*
 * Internal pager
 */
struct binding_t Op_pager[] = {
  { "next-page", OP_PAGER_NEXT_PAGE },
  { "previous-page", OP_PAGER_PREVIOUS_PAGE },
  { "next-line", OP_PAGER_NEXT_LINE },
  { "previous-line", OP_PAGER_PREVIOUS_LINE },
  { "top", OP_PAGER_TOP },
  { "bottom", OP_PAGER_BOTTOM },
  { "search", OP_PAGER_SEARCH },

  { "reply", OP_MAIN_REPLY },
  { "group-reply", OP_MAIN_GROUP_REPLY },
  { "next-message", OP_MAIN_NEXT_MESSAGE },
  { "previous-message", OP_MAIN_PREVIOUS_MESSAGE },
  { "bounce-message", OP_MAIN_BOUNCE_MESSAGE },
  { "display-headers", OP_MAIN_DISPLAY_HEADERS },
  { "exit", OP_PAGER_EXIT },
  { "delete-message", OP_MAIN_DELETE_MESSAGE },
  { "undelete-message", OP_MAIN_UNDELETE_MESSAGE },
  { "save-message", OP_MAIN_SAVE_MESSAGE },
  { "copy-message", OP_MAIN_COPY_MESSAGE },
  { "list-reply", OP_MAIN_LIST_REPLY },
  { "print-message", OP_MAIN_PRINT_MESSAGE },
  { "pipe-message", OP_MAIN_PIPE_MESSAGE },
  { "forward-message", OP_MAIN_FORWARD_MESSAGE },
  { "next-undeleted", OP_MAIN_NEXT_UNDELETED },
  { "previous-undeleted", OP_MAIN_PREVIOUS_UNDELETED },
  { 0, 0 }
};

/*
 * Main menu
 */
struct binding_t Op_main[] = {
  { "next-undeleted", OP_MAIN_NEXT_UNDELETED },
  { "next-message", OP_MAIN_NEXT_MESSAGE },
  { "previous-undeleted", OP_MAIN_PREVIOUS_UNDELETED },
  { "previous-message", OP_MAIN_PREVIOUS_MESSAGE },
  { "delete-message", OP_MAIN_DELETE_MESSAGE },
  { "undelete-message", OP_MAIN_UNDELETE_MESSAGE },
  { "delete-pattern", OP_MAIN_DELETE_PATTERN },
  { "undelete-pattern", OP_MAIN_UNDELETE_PATTERN },
  { "save-message", OP_MAIN_SAVE_MESSAGE },
  { "copy-message", OP_MAIN_COPY_MESSAGE },
  { "print-message", OP_MAIN_PRINT_MESSAGE },
  { "pipe-message", OP_MAIN_PIPE_MESSAGE },
  { "limit", OP_MAIN_LIMIT },
  { "tag-message", OP_MAIN_TAG_MESSAGE },
  { "tag-pattern", OP_MAIN_TAG_PATTERN },
  { "first-message", OP_MAIN_FIRST_MESSAGE },
  { "last-message", OP_MAIN_LAST_MESSAGE },
  { "quit-folder", OP_MAIN_QUIT },
  { "exit-folder", OP_MAIN_EXIT },
  { "sort-messages", OP_MAIN_SORT_MESSAGES },
  { "change-folder", OP_MAIN_CHANGE_FOLDER },
  { "view-attachments", OP_MAIN_VIEW_ATTACHMENTS },
  { "jump", OP_MAIN_JUMP },
  { "alias", OP_MAIN_ALIAS },
  { "bounce-message", OP_MAIN_BOUNCE_MESSAGE },
  { "next-page", OP_MAIN_NEXT_PAGE },
  { "previous-page", OP_MAIN_PREVIOUS_PAGE },
  { "display-message", OP_MAIN_DISPLAY_MESSAGE },
  { "display-headers", OP_MAIN_DISPLAY_HEADERS },
  { "sync-folder", OP_MAIN_SYNC_FOLDER },
  { "reply-message", OP_MAIN_REPLY },
  { "group-reply", OP_MAIN_GROUP_REPLY },
  { "search", OP_MAIN_SEARCH },
  { "forward-message", OP_MAIN_FORWARD_MESSAGE },
  { "help", OP_MAIN_HELP },
  { "refresh-screen", OP_MAIN_REFRESH },
  { "mail-message", OP_MAIN_MAIL },
  { "list-reply", OP_MAIN_LIST_REPLY },
  { "flag-message", OP_MAIN_FLAG_MESSAGE },
  { "mark-as-new", OP_MAIN_MARK_AS_NEW },
#ifdef _PGPPATH
  { "forget-passphrase", OP_MAIN_FORGET_PASSPHRASE },
#endif
  { 0, 0 }
};

/*
 * Send menu
 */
struct binding_t Op_send[] = {
  { "help-menu", OP_SEND_HELP_MENU },
  { "edit-to", OP_SEND_EDIT_TO },
  { "edit-cc", OP_SEND_EDIT_CC },
  { "edit-bcc", OP_SEND_EDIT_BCC },
  { "edit-message", OP_SEND_EDIT_MESSAGE },
  { "edit-headers", OP_SEND_EDIT_HEADERS },
  { "edit-subject", OP_SEND_EDIT_SUBJECT },
  { "send-message", OP_SEND_SEND_MESSAGE },
  { "abort-message", OP_SEND_ABORT_MESSAGE },
#ifdef ISPELL
  { "ispell", OP_SEND_ISPELL },
#endif
#ifdef MIXMASTER_PATH
  { "mixmaster-menu", OP_SEND_MIXMASTER_MENU },
#endif
#ifdef _PGPPATH
  { "pgp-menu", OP_SEND_PGP_MENU },
#endif /* PATHPATH */
  { "alias-menu", OP_SEND_ALIAS_MENU },
  { "attach-file", OP_SEND_ATTACH_FILE },
  { "refresh", OP_SEND_REFRESH },
  { "postpone-message", OP_SEND_POSTPONE_MESSAGE },
  { 0, 0 }
};
