/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#ifndef MUTTLIB_H
#define MUTTLIB_H

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <limits.h>
#include <setjmp.h>
#include "config.h"

#define T 1
#define NIL 0

#ifdef HAVE_SRAND48
#define LRAND lrand48
#define SRAND srand48
#define DRAND drand48
#else
#define LRAND rand
#define SRAND srand
#define DRAND (double)rand
#endif /* HAVE_SRAND48 */

#if defined(__hp9000s700) || defined(__hp9000s8000)
#define _HPUX_SOURCE 1
#endif

#define whitespace(c) ((c) == ' ' || (c) == '\t' || (c) == '\r' || (c) == '\n')

#define FOREVER while (1)

#define MAGICMBOX   1
#define MAGICMMDF   2

/* The stream corresponding to the currently openend folder */
#define MAILFP Context->folder->fp

/* Content-Disposition values */
#define DISPINLINE 0
#define DISPATTACH 1

#define TYPE(X) body_types[(X)]
#define ENCODING(X) body_encodings[(X)]

#if DEBUG
#define dprint(N,X) if(debuglevel>=N) fprintf X
#else
#define dprint(N,X) 
#endif

#define strfcpy(A,B,C) strncpy(A,B,C); A[(C)-1]=0;

#define LONG_STRING     1024
#define STRING          256
#define SHORT_STRING    128

/*
 * This is the maximum number of headers that a user can define when sending
 * a message.
 */
#define MAXUSERHDRS 25

/* Content-Type */
#define TYPEOTHER       0
#define TYPEAUDIO       1
#define TYPEAPPLICATION 2
#define TYPEIMAGE       3
#define TYPEMESSAGE     4
#define TYPEMULTIPART   5
#define TYPETEXT        6
#define TYPEVIDEO       7

/* Content-Transfer-Encoding */
#define ENCOTHER           0
#define ENC7BIT            1
#define ENC8BIT            2
#define ENCQUOTEDPRINTABLE 3
#define ENCBASE64          4
#define ENCBINARY          5

#define ACTIONLIMIT    1
#define ACTIONSELECT   2
#define ACTIONFIND     3
#define ACTIONUNSELECT 4
#define ACTIONDELETE   5
#define ACTIONUNDELETE 6

#define SENDREPLY      1
#define SENDGROUPREPLY 2
#define SENDLISTREPLY  4
#define SENDFORWARD    8
#define SENDBATCH      16

#define SORTDATE    0   /* the date the mail was sent. */
#define SORTSIZE    1
#define SORTSUBJECT 2
#define SORTFROM    3
#define SORTORDER   4   /* the order the messages appear in the mailbox. */
#define SORTMAX     SORTORDER /* FOO - needs to be changed if more methods */
#define SORTMASK    0x7
#define SORTREVERSE 8

#define OPTPROMPTAFTER 1
#define OPTKEEPSAVE    (1<<1)
#define OPTXTERMMOUSE  (1<<2)
#define OPTALLOW8BIT   (1<<3)
#define OPTASKCOPY     (1<<4)
#define OPTHOLD        (1<<5)
#define OPTMETOO       (1<<6)
#define OPTEDITHDRS    (1<<7)
#define OPTARROWCURSOR (1<<8)
#define OPTASKCC       (1<<9)
#define OPTHEADER      (1<<10)
#define OPTPOINTNEW    (1<<11)
#define OPTASKDEL      (1<<12)
#define OPTASKMOVE     (1<<13)
#define OPTREVALIAS    (1<<14)
#define OPTFORCENAME   (1<<15)
#define OPTKEEPEMPTY   (1<<16)
#define OPTMIMEFORWARD (1<<17)
#define OPTSHOWHOST    (1<<18)
#define OPTASKPOSTPONE (1<<19)
#define OPTASKREPLYTO  (1<<20)
#define OPTSIGDASHES   (1<<21)
#define OPTFLOATINGPROMPT (1<<22)
#define OPTHONORLENGTH (1<<23)
#define option(c) (Options & (c))

/* A dynamic array */
typedef struct {
  char **data;
  short len;
  short max;
} ARRAY;

typedef struct {
  FILE *fpin;
  FILE *fpout;
  char *prefix;
  int displaying; /* TRUE only if we are sending the output to a pager. */
  ARRAY *array;
  char buffer[1024];
  short offset;
} STATE;

typedef struct address ADDRESS;
struct address {
  char *personal;
  char *mailbox;
  char *adl;
  char *host;
  char *error;
  ADDRESS *next;
};

typedef struct alias {
  char *name;
  ADDRESS *addr;
  struct alias *next;
} ALIAS;

typedef struct list_t {
  char *data;
  struct list_t *next;
} LIST;

typedef struct envelope {
  ADDRESS *from;
  ADDRESS *to;
  ADDRESS *cc;
  ADDRESS *bcc;
  ADDRESS *sender;
  char *subject;
  char *in_reply_to;
  LIST *references;
} ENVELOPE;

typedef struct userheader {
  char *key;
  char *val;
  struct userheader *next;
} USERHEADER;

#define MAXUSERHDRS 25
typedef struct metaenv {
  ENVELOPE *env;
  char *userhdrs[MAXUSERHDRS];
} METAENV;

typedef struct parameter PARAMETER;
struct parameter {
  char *attribute;
  char *value;
  PARAMETER *next;
};

/* Information that helps in determing the Content-* of an attachment */
typedef struct content {
  unsigned int space : 1; /* whitespace at the end of lines? */
  unsigned int binary : 1; /* long lines, or CR not in CRLF pair */
  unsigned int from : 1; /* has a line beginning with "From "? */
  unsigned int dot : 1; /* has a line consisting of a single dot? */
  long hibin; /* 8-bit characters */
  long lobin; /* unprintable 7-bit chars (eg., control chars) */
  long ascii;
  long linemax; /* length of the longest line in the file */
} CONTENT;

typedef struct body {
  unsigned int type : 3;
  unsigned int encoding : 3;
  unsigned int disposition : 1;
  unsigned int use_disp : 1; /* Content-Disposition field printed? */
  unsigned int unlink : 1;
  char *subtype;
  PARAMETER *parameter;
  char *description;
  char *md5;
  long hdr_offset;
  long offset;
  long length;
  char *filename;
  CONTENT *content;
  struct body *next;
  struct body *parts; /* parts of a multipart message */
} BODY;

typedef struct header {
  unsigned int mime : 1; /* has a Mime-Version header? */
  unsigned int metamail : 1; /* requires metamail? */
  unsigned int flagged : 1;
  unsigned int visible : 1;
  unsigned int selected : 1;
  unsigned int deleted : 1;
  unsigned int changed : 1;
  unsigned int old : 1;

  unsigned int read : 1;
  unsigned int replied : 1;
  unsigned int marked : 1;
  unsigned int remail : 1; /* this message was bounced from someplace. */
#ifdef _PGPPATH
  unsigned int pgp : 2;
#endif

  /* Components from the "DATE:" field. */
  unsigned int year : 7;
  unsigned int month : 5;
  unsigned int day : 6;
  unsigned int hours : 5;
  unsigned int minutes : 6;
  unsigned int seconds : 6;
  unsigned int zhours : 5;
  unsigned int zminutes : 6;
  unsigned int zoccident : 1;

  time_t received;

  short msgno;
  short virtual;
  long offset; /* where in the stream does this message begin? */
  long length; /* how long is this message (headers + body)? */
  short lines; /* how many lines in the body of this message? */
  ADDRESS *from;
  ADDRESS *to;
  ADDRESS *cc;
  ADDRESS *sender;
  ADDRESS *reply_to;
  char *subject;
  char *message_id;
  LIST *references;

  BODY *content;
} HEADER;

typedef struct {
  char *name;
  short magic;
  FILE *fp;
  time_t mtime; /* last modification date */
  long size;    /* Size of the mailbox.  */
} FOLDER;

typedef struct {
  FOLDER *folder;
  HEADER **hdrs;
  short hdrmax;
  short **v2r; /* mapping from virtual to real msgno */
  short msgcount;
  short vcount; /* the number of virtual messages */
  short new;
  short current;
  short selected; /* one or more messages are selected */
  unsigned int weed : 1;
  unsigned int locked : 1;
} CONTEXT;

/*
 * Function prototypes.
 */

FOLDER *folder_open(const char *, const char *mode);
void folder_close(FOLDER **);

int lock_file(FOLDER *, int);
void unlock_file(FOLDER *);

ADDRESS *mutt_default_from (void);
ADDRESS *mutt_expand_aliases (ADDRESS *);
ADDRESS *mutt_new_address (void);
BODY *mutt_new_body (void);
BODY *mutt_read_mime_header (FILE *);
BODY *mutt_make_attach (const char *);
BODY *mutt_make_multipart (BODY *, BODY *);
CONTENT *mutt_get_content_info (const char *);
ENVELOPE *mutt_new_envelope (void);
HEADER *mutt_new_header (void);
METAENV *mutt_new_metaenv (void);
PARAMETER *mutt_new_parameter (void);
STATE *mutt_new_state (void);

char *mutt_make_date (char *);
char *mutt_new_tempfile (void);
char *mutt_get_parameter (const char *, PARAMETER *);
char *mutt_generate_boundary (void);

int mutt_addr_is_user (ADDRESS *);
int mutt_is_mail_list (ADDRESS *);

CONTEXT *mutt_open_mailbox (char *, CONTEXT *);
int mutt_check_mailbox (CONTEXT *);
int mutt_parse_mailbox (CONTEXT *);
int mutt_sync_mailbox (CONTEXT *);
void mutt_close_mailbox (CONTEXT *);
void mutt_sort_headers (CONTEXT *);
void mutt_fastclose_mailbox (CONTEXT *);

void mutt_add_to_list (LIST **, char *);
void mutt_body_handler (BODY *, STATE *);
void mutt_bounce_message (HEADER *, char *);
void mutt_complete (char *);
void mutt_copy_bytes (FILE *, FILE *, long);
void mutt_copy_header (HEADER *, FILE *, int);
void mutt_decode_base64 (STATE *, long, int);
void mutt_decode_quoted (STATE *, long, int);
void mutt_decode_xbit (STATE *, long);
void mutt_default_savename (char *, HEADER *);
void mutt_error (const char *, ...);
void mutt_exit (int);
void mutt_expand_path (char *);
void mutt_free_address (ADDRESS **);
void mutt_free_body (BODY **);
void mutt_free_envelope (ENVELOPE **);
void mutt_free_header (HEADER **);
void mutt_free_state (STATE **);
void mutt_free_parameter (PARAMETER **);
void mutt_generate_header (char *, HEADER *, int);
void mutt_global_init (void);
void mutt_make_string (char *, char *, HEADER *, int);
void mutt_parse_mime_message (HEADER *);
void mutt_parse_references (LIST **, char *);
void mutt_postpone_message (const char *, METAENV *, const char *);
void mutt_pipe_message (HEADER *, const char *);
void mutt_set_content_info (BODY *);
void mutt_show_version (void);
void mutt_tabs_to_spaces (char *);
void mutt_text_handler (BODY *, STATE *);

int mutt_create_filter (const char *, FILE **, FILE **);
int mutt_init (int);
int mutt_is_text_type (int, char *);
int mutt_matches_ignore (const char *);
int mutt_needs_metamail (BODY *);
int mutt_parse_bind (char *);
int mutt_print_message (HEADER *);
int mutt_save_attachment (BODY *, char *);
int mutt_save_message (HEADER *, char *, int);
int mutt_send_message (METAENV *, BODY *, int);
int mutt_system (const char *);
int mutt_write_mime_body (BODY *, FILE *);
int mutt_write_mime_header (BODY *, FILE *);
int mutt_write_rfc822_header (FILE *, METAENV *, BODY *, int);
int mutt_yesorno (const char *, int);

int state_puts (char *, STATE *);
int state_putc (int, STATE *);

char *safe_strdup (const char *);
void *safe_malloc (unsigned int);
void safe_realloc (void **, unsigned int);
void safe_free (void **);

void save_to_folder (const char *, int, int);
void append_file_to_folder (const char *, char *);
void copy_header (STATE *, int);


/* from.c */
time_t is_from (char *);

int compat_charset (const char *);
char *skip_whitespace (char *);

/* aliaslib.c */
ADDRESS *alias_reverse_lookup (ADDRESS *);

/* parse.c */
BODY *parse_multipart (FILE *, char *);
BODY *parse_messageRFC822 (FILE *, BODY *);
BODY *read_mime_header (FILE *);

/* pgp.c */
#ifdef _PGPPATH
#define PGPENCRYPT 1
#define PGPSIGN    2
#define PGPKEY     3

void application_pgp_handler (BODY *, STATE *);
void pgp_signed_handler (BODY *, STATE *);
void pgp_encrypted_handler (BODY *, STATE *);
BODY *pgp_signMessage (BODY *);
BODY *pgp_encryptMessage (BODY *, char *, int);
char *pgp_findKeys (ADDRESS *, ADDRESS *, ADDRESS *);
int pgp_sendMenu (void);
int pgp_query (BODY *);
void pgp_void_passphrase (void);
#endif /* _PGPPATH */

extern int strcasecmp (const char *, const char *);
extern int strncasecmp (const char *, const char *, size_t);

extern FILE *fdopen (int, const char *);
extern int system (const char *);

#ifndef STDC_HEADERS
#include <stdlib.h>
#include <varargs.h>

extern int puts();
extern int fputs();
extern int fputc();
extern int fseek();
extern char *strchr();
extern int getopt();
extern int fputs();
extern int fputc();
extern int fclose();
extern int fprintf();
extern int printf();
extern int fgetc();
extern int tolower();
extern int toupper();
extern int sscanf();
extern size_t fread();
extern size_t fwrite();
extern int system();
extern int rename();
extern time_t time();
extern struct tm *localtime();
extern char *asctime();
extern char *strpbrk();
extern int fflush();
extern int ioctl();
extern long lrand48();
extern void srand48();
extern time_t mktime();
extern int vsprintf(char *, const char *, va_list);
extern int ungetc();
extern char *mktemp(char *);
extern int ftruncate(int, off_t);
extern char *re_comp(const char *);
extern int re_exec(const char *);
#endif

/*
 * Global variable definitions
 */

/* nifty trick I stole from ELM 2.5alpha. */
#ifdef MAIN_C
#define WHERE 
#define INITVAL(x) = x
#else
#define WHERE extern
#define INITVAL(x) 
#endif

WHERE jmp_buf Jumpbuf;

WHERE CONTEXT *Context;

WHERE char Errorbuf[SHORT_STRING];

WHERE char Attribution[SHORT_STRING];
WHERE char Charset[SHORT_STRING];
WHERE char Domain[SHORT_STRING];
WHERE char Editor[SHORT_STRING];
WHERE char Fqdn[SHORT_STRING];
WHERE char Hdr_format[SHORT_STRING];
WHERE char Homedir[_POSIX_PATH_MAX];
WHERE char Hostname[SHORT_STRING];
WHERE char In_reply_to[SHORT_STRING];
WHERE char Inbox[_POSIX_PATH_MAX];
WHERE char Local_sig[_POSIX_PATH_MAX];
WHERE char Maildir[_POSIX_PATH_MAX];
WHERE char Metamail[_POSIX_PATH_MAX];
WHERE char Outbox[_POSIX_PATH_MAX];
WHERE char Pager[_POSIX_PATH_MAX];
WHERE char Pgp[_POSIX_PATH_MAX];
WHERE char PostIndentString[SHORT_STRING];
WHERE char Postponed[_POSIX_PATH_MAX];
WHERE char Prefix[SHORT_STRING];
WHERE char Print[_POSIX_PATH_MAX];
WHERE char Realname[SHORT_STRING];
WHERE char Remote_sig[_POSIX_PATH_MAX];
WHERE char Sendmail[_POSIX_PATH_MAX];
WHERE char Signature[_POSIX_PATH_MAX];
WHERE char Spoolfile[_POSIX_PATH_MAX];
WHERE char Tempdir[_POSIX_PATH_MAX];
WHERE char Username[SHORT_STRING];

WHERE LIST *Ignore INITVAL(0);
WHERE LIST *Mailing_lists INITVAL(0);

WHERE long Options INITVAL(OPTKEEPSAVE|OPTPROMPTAFTER|OPTHOLD|OPTALLOW8BIT|OPTPOINTNEW|OPTASKDEL|OPTASKMOVE|OPTASKPOSTPONE|OPTSIGDASHES|OPTFLOATINGPROMPT);

WHERE int Default_magic INITVAL(MAGICMBOX);
WHERE int Msg_inc INITVAL(10);
WHERE int Sleep_time INITVAL(1);
WHERE int Sort INITVAL(SORTDATE);

WHERE ALIAS *Aliases INITVAL(0);
WHERE ADDRESS *Alternates INITVAL(0);
WHERE USERHEADER *User_defined_headers INITVAL(0);

#if DEBUG
WHERE FILE *debugfile INITVAL(0);
WHERE int debuglevel INITVAL(0);
#endif

#ifdef USE_SETGID
WHERE gid_t Mail_gid;
WHERE gid_t User_gid;
#endif /* USE_SETGID */

WHERE char *Mutt_version;

#ifdef MAIN_C
const char *Weekdays[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
const char *Months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec", "ERR" };

const char *body_types[] = { "x-unknown", "audio", "application", "image", "message", "multipart", "text", "video" };
const char *body_encodings[] = { "x-unknown", "7bit", "8bit", "quoted-printable", "base64", "binary" };
#else
extern const char *Weekdays[];
extern const char *Months[];

extern const char *body_types[];
extern const char *body_encodings[];
#endif /* GLOBAL_C */

#endif /* MUTTLIB_H */
