/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "mutt.h"
#include "rfc822.h"

#define BCC     1
#define CC      2
#define FROM    3
#define SUBJECT 4
#define TO      5

/* Returns true if string "a" is found in string "b" */
static int pattern_match (char *a, char *b) {
  int len = strlen(a);

  while (*b) {
    if (strncasecmp(a, b, len) == 0) return 1;
    b++;
  }
  return 0;
}

static int meta_match (int f, HEADER *h, char *c) {
  char buffer[STRING];
  ADDRESS *a;

  switch (f) {
  case SUBJECT:
    return (h->subject ? pattern_match(c, h->subject) : 0);
  case FROM:
    a = h->from;
    break;
  case TO:
    a = h->to;
    break;
  case CC:
    a = h->cc;
    break;
  default:
    return 0;
  }
  buffer[0] = 0;
  rfc822_write_address(buffer, a);
  return(pattern_match(c, buffer));
}

/*
 * A generic function to delete/undelete/tag/untag/limit/unlimit messages
 * which match a patter inputted by the user.
 */
void pattern_function (int op, const char *prompt) {
  int field = 0, i;
  char *crit, buffer[256];
  int matched = 0;

  buffer[0] = 0;
  ci_get_field(prompt, buffer, 0, 0);
  if (!buffer[0]) return;
  if (strcasecmp(buffer, "all") == 0) {
    for (i = 0; i < Context->msgcount; i++) {
      switch (op) {
      case ACTIONDELETE:
        Context->hdrs[i]->deleted = 1;
        break;
      case ACTIONUNDELETE:
        Context->hdrs[i]->deleted = 0;
        break;
      case ACTIONSELECT:
        Context->hdrs[i]->selected = 1;
        break;
      case ACTIONLIMIT:
    Context->hdrs[i]->visible = 1;
	Context->hdrs[i]->virtual = i;
	*Context->v2r[i] = i;
	break;
      }
    }
    if (op == ACTIONSELECT)
      Context->selected = Context->msgcount;
    else if (op == ACTIONLIMIT)
      Context->vcount = Context->msgcount;
    return;
  }
  else if (op == ACTIONSELECT && strcmp(buffer, "none") == 0) {
    for (i=0; i<Context->msgcount; i++)
      Context->hdrs[i]->selected = 0;
    Context->selected = 0;
    return;
  }
  else if (strcmp(buffer, "tagged") == 0) {
    if (op == ACTIONLIMIT) Context->vcount = 0;
    for (i = 0; i < Context->msgcount; i++) {
      if (Context->hdrs[i]->selected) {
	switch (op) {
	case ACTIONDELETE:
	  Context->hdrs[i]->deleted = 1;
	  break;
	case ACTIONUNDELETE:
	  Context->hdrs[i]->deleted = 0;
	  break;
	case ACTIONLIMIT:
      Context->hdrs[i]->visible = 1;
      matched++;
	  break;
	default:
	  break;
	}
      }
      else if (op == ACTIONLIMIT)
    Context->hdrs[i]->visible = 0;
    }

    if (op == ACTIONLIMIT)
      goto reset_virtual;
    return;
  }

  crit = buffer;
  while ((crit = strtok(crit, " \t")) != NULL) {
    if (tolower(*crit) == 'f') {
      if (strcasecmp("rom", crit+1) == 0) {
        field = FROM;
      }
    }
    else if (tolower(*crit) == 's') {
      if (strcasecmp("ubject", crit+1) == 0 || strcasecmp("ubj", crit+1) == 0)
        field = SUBJECT;
    }
    else if (tolower(*crit) == 'c') {
      if (strcasecmp("c", crit+1) == 0)
        field = CC;
    }
    else if (tolower(*crit) == 't') {
      if (strcasecmp("o", crit+1) == 0) {
        field = TO;
      }
    }
    else
      return;

    if ((crit = strtok(NULL, " \t")) == NULL) {
      mutt_error("Missing argument!");
      return;
    }

    i = (op == ACTIONFIND ? Context->current+1 : 0);
    for (; i < Context->msgcount; i++) {
      if (meta_match(field, Context->hdrs[i], crit)) {
        if (op == ACTIONLIMIT) {
           Context->hdrs[i]->visible = 1;
	       matched++;
   	    }
        else if (op == ACTIONSELECT) {
          if (! Context->hdrs[i]->selected) {
            Context->hdrs[i]->selected = 1;
            Context->selected++;
          }
        }
        else if (op == ACTIONFIND) {
	  Context->current = i;
	  return;
	}
	else if (op == ACTIONDELETE)
	  Context->hdrs[i]->deleted = 1;
	else if (op == ACTIONUNDELETE)
	  Context->hdrs[i]->deleted = 0;
      }
      else {
        if (op == ACTIONLIMIT)
          Context->hdrs[i]->visible = 0;
        else if (op == ACTIONSELECT) {
          if (! Context->hdrs[i]->selected) {
            Context->hdrs[i]->selected = 0;
            Context->selected--;
          }
        }
      }
    }
    crit = 0;
  }

  if (op == ACTIONLIMIT) {
reset_virtual:
    if (matched) {
      Context->vcount = 0;
      for (i=0; i < Context->msgcount; i++) {
   if (Context->hdrs[i]->visible) {
     *Context->v2r[Context->vcount] = i;
     Context->hdrs[i]->virtual = Context->vcount;
     Context->vcount++;
   }
   else
     Context->hdrs[i]->virtual = -1;
      }
      Context->current = 0;
    }
    else {
      mutt_error("No messages matched!");
      
      /* reset the virtual message numbers */
      for (i = 0; i < Context->msgcount; i++) {
   *Context->v2r[i] = i;
   Context->hdrs[i]->virtual = i;
      }
      Context->vcount = Context->msgcount;
    }
  }
  else if (op == ACTIONFIND)
    mutt_error("Not found!");
}
