/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

/* WARNING!!!  WARNING!!!  WARNING!!!  WARNING!!!  WARNING!!!  WARNING!!!
 *
 * Legal for distribution in the U.S./Canada ONLY!  Distributing this file
 * outside of the U.S./Canada may be in violation of ITAR regulations!
 */

/* This file contains all of the PGP routines necessary to sign, encrypt,
 * verify and decrypt PGP messages in either the new PGP/MIME format, or
 * in the older Application/Pgp format.  It also contains some code to
 * cache the user's passphrase for repeat use when decrypting or signing
 * a message.
 */

#include <sys/wait.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <errno.h>
#include "mutt.h"
#include "mutt_curses.h"
#include "rfc822.h"

#ifdef _PGPPATH

struct keyinfo {
  char *keyid;
  ADDRESS *address;
  struct keyinfo *next;
};
typedef struct keyinfo KEYINFO;

static char pgp_passphrase[256];
static time_t pgp_exptime; /* when does the cached passphrase expire? */
static time_t pgp_timeout = 300; /* default to 5 minutes. */

void pgp_void_passphrase (void)
{
  int i;

  for (i = 0; i < sizeof(pgp_passphrase); i++)
    pgp_passphrase[i] = 0;
}

static int pgp_prompt_for_passphrase (void)
{
  mvaddstr(LINES-2, 0, "Please enter your PGP passphrase: ");
  ci_enter_string(pgp_passphrase, LINES-2, sizeof("Please enter your PGP passphrase: ")-1, 1, 0);
  ClearLine(LINES-2);
  if (pgp_passphrase[0]) {
    pgp_exptime = time(0) + pgp_timeout;
    return 1;
  }
  pgp_exptime = 0;
  return 0;
}

static int pgp_valid_passphrase (void)
{
  time_t now = time(0);
  if (now < pgp_exptime) return 1; /* just use the cached copy. */
  pgp_void_passphrase();
  return(pgp_prompt_for_passphrase());
}

/*
 * The following routines are used to parse the output of "pgp -kv" to build
 * a list of the public keys in the user's database.
 */
static KEYINFO *ki_parse_line (char *s)
{
  char *keyid, *address;
  KEYINFO *p;

  if ((s = strchr(s, '/')) == 0) return 0;
  s++;
  keyid = s;
  while (*s && *s != ' ') s++;
  if (!*s) return 0;
  *s++ = 0;
  while (*s && isspace(*s)) s++;
  if ((s = strchr(s, ' ')) == 0) return 0;
  while (*s && isspace(*s)) s++;
  address = s;
  while (*s && *s != '\n') s++;
  *s = 0;

  p = (KEYINFO *)malloc(sizeof(KEYINFO));
  p->keyid = strdup(keyid);
  p->address = 0;
  p->next = 0;
  rfc822_parse_adrlist(&p->address, address, Hostname);
  return(p);
}

static KEYINFO *ki_getdb (void)
{
  FILE *in;
  KEYINFO *db = NULL, *end = 0;
  char buffer[STRING], *c;

  sprintf(buffer, "%s +verbose=0 -kv", Pgp);
  mutt_create_filter(buffer, 0, &in);
  while (fgets(buffer, STRING, in) != NULL) {
    if (strncmp(buffer, "pub", 3) == 0 || strncmp(buffer, "sec", 3) == 0) {
      if (db) {
        end->next = ki_parse_line(buffer);
        end = end->next;
      }
      else
        db = end = ki_parse_line(buffer);
    }
    else if (buffer[0] == ' ') { /* alias */
      c = buffer;
      while (*c && (*c == ' ' || *c == '\t')) c++;
      rfc822_parse_adrlist(&end->address, c, Hostname);
    }
  }
  fclose(in);
  return(db);
}

static void ki_closedb (KEYINFO *k)
{
  KEYINFO *tmp;

  while (k) {
    if (k->keyid) free(k->keyid);
    mutt_free_address(&k->address);
    tmp = k;
    k = k->next;
    free(tmp);
  }
}

static char *ki_getkeyidbyaddr (ADDRESS *a, KEYINFO *k)
{
  ADDRESS *p;

  while (k) {
    p = k->address;
    while (p) {
      if (p->mailbox && p->host && a->mailbox && a->host &&
          strcasecmp(p->mailbox, a->mailbox) == 0 && 
          strcasecmp(p->host, a->host) == 0)
        return k->keyid;
      if (a->personal && p->personal && strcasecmp(a->personal, p->personal) == 0)
        return k->keyid;
      p = p->next;
    }
    k = k->next;
  }
  return 0;
}

/****************************************************************************/
/* Support for the Application/PGP Content Type.  This will fail if the CTE
 * is something other than 7BIT or 8BIT, but I've never seen a message like
 * that.  Hopefully more people will start using Multipart/Signed and
 * Multipart/Encrypted instead, because it's much more robust.
 */
void application_pgp_handler (BODY *m, STATE *s)
{
  int status, needpass = -1;
  long len, bytes;
  char buffer[LONG_STRING];
  FILE *pgpout, *pgpin;

  fseek(s->fpin, m->offset, 0);
  for (bytes = m->length; bytes > 0;) {
    if (fgets(buffer, sizeof(buffer)-1, s->fpin) == NULL) return;
    if (strncmp("-----BEGIN PGP ", buffer, 15) == 0) {
      if (strcmp("MESSAGE-----\n", buffer+15) == 0)
        needpass = 1;
      else
        needpass = 0;
      break;
    }
  }
  if (needpass < 0) return; /* not a PGP message! */
  else if (needpass) {
    if (!pgp_valid_passphrase()) {
      state_puts("[Error!  No valid passphrase.]\n", s);
      return;
    }
  }

  if (s->displaying) {
    if (needpass)
      state_puts("[-----BEGIN PGP MESSAGE-----]\n\n", s);
    else
      state_puts("[-----BEGIN PGP SIGNED MESSAGE-----]\n\n", s);
  }
  endwin();
  if (needpass)
    strcpy(buffer, "PGPPASSFD=0; export PGPPASSFD; ");
  else
    buffer[0] = 0;
  sprintf(buffer+strlen(buffer), "%s +verbose=0 +batchmode -f", Pgp);
  mutt_create_filter(buffer, &pgpin, &pgpout);
  fseek(s->fpin, m->offset, 0);
  
  if (needpass) {
    fputs(pgp_passphrase, pgpin);
    fputc('\n', pgpin);
  }
  mutt_copy_bytes(s->fpin, pgpin, m->length);
  fclose(pgpin);
  if (s->prefix)
    while (fgets(buffer, sizeof(buffer), pgpout) != NULL) {
      state_puts(s->prefix, s);
      state_puts(buffer, s);
    }
  else {
    /*
     * Since we don't need to prepend a string to each line, just copy the
     * stream in as big chunks as we can handler.  This will speed things
     * up somewhat.
     */
    while ((len = fread(buffer, 1, sizeof(buffer), pgpout)) > 0)
      fwrite(buffer, 1, len, s->fpout);
  }
  fclose(pgpout);
  wait(&status);
  if (s->displaying) {
    if (needpass) {
      state_puts("[-----END PGP MESSAGE-----]\n", s);
    }
    else {
      state_puts("[-----END PGP SIGNED MESSAGE-----]\n", s);
    }
  }
  /*
   * Pause to allow the user to read any PGP error messages.
   */
  ci_any_key_to_continue(0);
}

int pgp_query (BODY *m)
{
  char *p;

  /* Check for old-style APPLICATION/PGP messages */
  if (m->type == TYPEAPPLICATION && strcasecmp(m->subtype, "pgp") == 0) {
    if ((p = mutt_get_parameter("x-action", m->parameter))) {
      if (strcasecmp(p, "sign") == 0) return PGPSIGN;
      if (strncasecmp(p, "encrypt", 7) == 0) return PGPENCRYPT;
    }
    return PGPENCRYPT;  /* not necessarily correct, but... */
  }
  /* Check for PGP/MIME messages. */
  if (m->type == TYPEMULTIPART) {
    if (strcasecmp(m->subtype, "signed") == 0) {
      if ((p = mutt_get_parameter("protocol", m->parameter)) &&
         strcasecmp(p, "application/pgp-signature") == 0)
        return PGPSIGN;
    }
    else if (strcasecmp(m->subtype, "encrypted") == 0) {
      if ((p = mutt_get_parameter("protocol", m->parameter)) &&
          strcasecmp(p, "application/pgp-encrypted") == 0)
        return PGPENCRYPT;
    }
  }
  return 0;
}

/*
 * This routine verfies a PGP/MIME signed body.
 */
void pgp_signed_handler (BODY *a, STATE *s)
{
  FILE *fp;
  char *tempfile, sigfile[STRING], buffer[LONG_STRING];
  int bytes, len;

  a = a->parts;

  /* First do some error checking to make sure that this looks like a valid
   * multipart/signed body.
   */
  if (a && a->next && a->next->type == TYPEAPPLICATION && a->next->subtype &&
      strcasecmp(a->next->subtype, "pgp-signature") == 0) {
    tempfile = mutt_new_tempfile();
    fp = fopen(tempfile, "w");
    
    fseek(s->fpin, a->hdr_offset, 0);
    bytes = a->length + a->offset - a->hdr_offset;
    while (bytes > 0) {
      fgets(buffer, sizeof(buffer)-1, s->fpin);
      len = strlen(buffer);
      /* Convert LF=>CRLF when required (probably always) */
      if (buffer[len-2] != '\r') {
        buffer[len-1] = '\r';
        buffer[len] = '\n';
        buffer[len+1] = 0;
      }
      fputs(buffer, fp);
      bytes -= len;
    }
    fclose(fp);
    
    /* Now grab the signature.  Since signature data is required to be 7bit,
     * we don't have to worry about doing CTE decoding...
     */
    sprintf(sigfile, "%s.asc", tempfile);
    fp = fopen (sigfile, "w");
    fseek (s->fpin, a->next->offset, 0);
    mutt_copy_bytes(s->fpin, fp, a->next->length);
    fclose(fp);
    
    /* Now we're ready to check the signature */
    endwin();
    sprintf(buffer, "%s +batchmode +verbose=0 %s", Pgp, tempfile);
    printf("Executing: %s\n", buffer);
    mutt_system(buffer);
    unlink(tempfile);
    safe_free((void **)&tempfile);
    unlink(sigfile);
    ci_any_key_to_continue (0);

    /* Now display the signed body */
    if (s->displaying)
      state_puts ("[The following data is PGP/MIME signed.]\n\n", s);
    mutt_body_handler (a, s);
    if (s->displaying)
      state_puts ("\n[End of PGP/MIME signed data.]\n", s);
  }
  else {
    dprint (1,(debugfile, "pgp_signed_handler: invalid format!\n"));
    state_puts ("[Error!  This message does not comply with the PGP/MIME specification!]\n\n", s);
    mutt_text_handler (a, s); /* just treat the data as text/plain... */
  }
}

void pgp_encrypted_handler (BODY *a, STATE *s)
{
  char *tempfile, buffer[2048];
  FILE *pgpin, *pgpout, *fpout;
  BODY *tattach;
  int usepass, len;
  struct stat info;

  a = a->parts;
  if (!a || a->type != TYPEAPPLICATION || !a->subtype || 
      strcasecmp("pgp-encrypted", a->subtype) != 0 ||
      !a->next || a->next->type != TYPEAPPLICATION || !a->next->subtype ||
      strcasecmp("octet-stream", a->next->subtype) != 0) {
    state_puts("[Error!  Malformed PGP/MIME message.  Could not process it!]\n\n", s);
    return;
  }

  if ((usepass = pgp_valid_passphrase()))
    strcpy(buffer, "PGPPASSFD=0; export PGPPASSFD; "); 
  else
    buffer[0] = 0;
  sprintf(buffer+strlen(buffer), "%s +verbose=0 +batchmode -f", Pgp);
  endwin();
  printf("Executing: %s\n", buffer);
  fflush(stdout);
  if (mutt_create_filter(buffer, &pgpin, &pgpout) == -1) {
    state_puts("[Error!  Could not create a PGP subprocess!]\n\n", s);
    return;
  }
  if (usepass) {
    fputs(pgp_passphrase, pgpin);
    fputc('\n', pgpin);
  }

  /*
   * Move forward to the application/pgp-encrypted body.
   */
  a = a->next;

  /*
   * Position the stream at the beginning of the body, and send the data to
   * the PGP subprocess.
   */
  fseek(s->fpin, a->offset, 0);
  mutt_copy_bytes(s->fpin, pgpin, a->length);
  fclose(pgpin);

  tempfile = mutt_new_tempfile();
  fpout = fopen(tempfile, "w");

  /*
   * Read the output from PGP, and make sure to change CRLF to LF, otherwise
   * read_mime_header has a hard time parsing the message.
   */
  while (fgets(buffer, sizeof(buffer)-1, pgpout) != NULL) {
    len = strlen(buffer);
    if (buffer[len-2] == '\r')
      strcpy(buffer+len-2, "\n");
    fputs(buffer, fpout);
  }

  fclose(pgpout);
  fclose(fpout);

  fpout = s->fpin;
  s->fpin = fopen(tempfile, "r");
  /*
   * unlink the file now so it disappears.  we can still read it, and it
   * will provide some amount of security for the decrypted data.
   */
  unlink(tempfile);
  safe_free((void **)&tempfile);

  /* allow the user to read the PGP output in case there was output. */
  ci_any_key_to_continue(0);

  tattach = mutt_read_mime_header(s->fpin);

  /*
   * Need to set the length of this body part.
   */
  fstat(fileno(s->fpin), &info);
  tattach->length = info.st_size - tattach->offset;

  /*
   * See if we need to recurse on this MIME part.
   */
  if (tattach->type == TYPEMULTIPART) {
    fseek(s->fpin, tattach->offset, 0);
    tattach->parts = parse_multipart(s->fpin, mutt_get_parameter("boundary", tattach->parameter));
  }
  else if (tattach->type == TYPEMESSAGE) {
    fseek(s->fpin, tattach->offset, 0);
    tattach->parts = parse_messageRFC822(s->fpin, tattach);
  }

  state_puts("[The follwing data is PGP/MIME encrypted.]\n\n", s);
  mutt_body_handler(tattach, s);
  state_puts("[End of PGP/MIME encrypted data.]\n", s);
  mutt_free_body(&tattach);
  fclose(s->fpin);
  s->fpin = fpout;
}

/*****************************************************************************
 * Routines for sending PGP/MIME messages.
 *****************************************************************************/

static void convert_to_7bit (BODY *a)
{
  while (a) {
    if (a->type == TYPEMULTIPART) {
      if (a->encoding != ENC7BIT) {
        a->encoding = ENC7BIT;
        convert_to_7bit (a->parts);
      }
    }
    else if (a->encoding == ENC8BIT)
      a->encoding = ENCQUOTEDPRINTABLE;
    else if (a->encoding == ENCBINARY)
      a->encoding = ENCBASE64;
    else if (a->content->from)
      a->encoding = ENCQUOTEDPRINTABLE;
    a = a->next;
  }
}

BODY *pgp_signMessage (BODY *a)
{
  PARAMETER *p;
  BODY *t;
  char cmd[STRING], *sigfile, buffer[LONG_STRING];
  FILE *pgpin, *pgpout, *fp;
  int usepass = 0, err = 0;

  convert_to_7bit (a); /* Signed data _must_ be in 7-bit format. */

  sigfile = mutt_new_tempfile();
  if ((fp = fopen(sigfile, "w")) == NULL) {
    safe_free((void **)&sigfile);
    return 0;
  }

#ifdef HAVE_PGPPASSFD
  if ((usepass = pgp_valid_passphrase()))
    strcpy(cmd, "PGPPASSFD=0; export PGPPASSFD; ");
  else
#endif
    cmd[0] = 0;
  sprintf(cmd+strlen(cmd), "%s +verbose=0 -abfst", Pgp);

  endwin();
  printf ("Executing: %s\n", cmd);
  fflush (stdout);

  if (mutt_create_filter(cmd, &pgpin, &pgpout) == -1) {
    safe_free((void **)&sigfile);
    fclose(fp);
    return 0;
  }

  if (usepass) {
    fputs (pgp_passphrase, pgpin);
    fputc ('\n', pgpin);
  }

  mutt_write_mime_header(a, pgpin);
  fputc('\n', pgpin);
  mutt_write_mime_body(a, pgpin);

  if (fclose (pgpin) != 0) {
    dprint(1,(debugfile, "pgp_sign_message: fclose() returned -1; errno=%d.\n",
              errno));
    fclose (pgpin);
    fclose (pgpout);
    fclose (fp);
    unlink (sigfile);
    safe_free((void **)&sigfile);
    return 0;
  }

  /* Read back the PGP signature.  Also, change MESSAGE=>SIGNATURE as
   * recommended for future releases of PGP.
   */
  while (fgets(buffer, sizeof(buffer)-1, pgpout) != NULL) {
    if (strcmp("-----BEGIN PGP MESSAGE-----\n", buffer) == 0)
      fputs("-----BEGIN PGP SIGNATURE-----\n", fp);
    else if (strcmp("-----END PGP MESSAGE-----\n", buffer) == 0)
      fputs("-----END PGP SIGNATURE-----\n", fp);
    else
      fputs(buffer, fp);
  }

  err = fclose(pgpout);
  if (fclose(fp) != 0 || err != 0) {
    unlink(sigfile);
    safe_free((void **)&sigfile);
    return 0;
  }

  ci_any_key_to_continue (0);

  t = mutt_new_body();
  t->type = TYPEMULTIPART;
  t->subtype = safe_strdup("signed");
  t->use_disp = 0;
  t->encoding = ENC7BIT;

  t->parameter = p = mutt_new_parameter();
  p->attribute = safe_strdup("protocol");
  p->value = safe_strdup("application/pgp-signature");

  p->next = mutt_new_parameter();
  p = p->next;
  p->attribute = safe_strdup("micalg");
  p->value = safe_strdup("pgp-md5");

  p->next = mutt_new_parameter();
  p = p->next;
  p->attribute = safe_strdup("boundary");
  p->value = mutt_generate_boundary();

  t->parts = a;
  a = t;

  t->parts->next = mutt_new_body();
  t = t->parts->next;
  t->type = TYPEAPPLICATION;
  t->subtype = safe_strdup("pgp-signature");
  t->filename = sigfile;
  t->use_disp = 0;
  t->encoding = ENC7BIT;
  t->unlink = 1; /* ok to remove this file after sending. */

  return(a);
}

/*
 * This routine attempts to find the keyids of the recipients of a message.
 * It returns NULL if any of the keys can not be found.
 */
char *pgp_findKeys (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc)
{
  KEYINFO *db = ki_getdb ();
  char *key, keylist[1024];
  ADDRESS *p = to;
  int count = 0;

  *keylist = 0;
  for (count = 0; count < 3; count++) {
    switch (count) {
    case 0:
      p = to;
      break;
    case 1:
      p = cc;
      break;
    case 2:
      p = bcc;
      break;
    }
    while (p) {
      key = ki_getkeyidbyaddr(p, db);
      if (!key) {
        ki_closedb(db);
        mutt_error("Could not find key for \"%s@%s\"", p->mailbox, p->host);
        return 0;
      }
      sprintf (keylist+strlen(keylist), " 0x%s", key);
      p = p->next;
    }
  }
  ki_closedb(db);
  return (strdup(keylist));
}

BODY *pgp_encryptMessage (BODY *a, char *keylist, int sign)
{
  char cmd[STRING], *tempfile, buffer[LONG_STRING];
  FILE *pgpin, *pgpout, *fpout;
  int len;
  BODY *t;
  PARAMETER *p;

  if (sign && (pgp_valid_passphrase() == 0)) return 0;
  endwin();
  if (sign) convert_to_7bit(a);
  sprintf(cmd, "%s%s +verbose=0 +encrypttoself -aef%s %s",
          sign ? "PGPPASSFD=0; export PGPPASSFD; " : "",
	  Pgp, sign ? "s" : "", keylist);
  if (mutt_create_filter(cmd, &pgpin, &pgpout) == -1) return 0;

  if (sign)
    {
      fputs(pgp_passphrase, pgpin);
      fputc('\n', pgpin);
    }

  mutt_write_mime_header(a, pgpin);
  fputc('\n', pgpin);
  mutt_write_mime_body(a, pgpin);
  if (fclose(pgpin) != 0) {
    dprint(1,(debugfile,"pgp_encrypt_message: fclose() returned -1; errno=%d.\n",
              errno));
    fclose(pgpin);
    fclose(pgpout);
  }
  tempfile = mutt_new_tempfile();
  if ((fpout = fopen(tempfile, "w")) == NULL) {
    dprint(1,(debugfile,"pgp_encrypt_mesage: error opening %s for writing!\n",
              tempfile));
    fclose(pgpin);
    fclose(pgpout);
    return 0;
  }
  while ((len = fread(buffer, 1, sizeof(buffer), pgpout)) > 0)
    fwrite(buffer, 1, len, fpout);
  fclose(pgpout);
  fclose(fpout);

  t = mutt_new_body();
  t->type = TYPEMULTIPART;
  t->subtype = safe_strdup("encrypted");
  t->encoding = ENC7BIT;
  t->use_disp = 0;

  t->parameter = p = mutt_new_parameter();
  p->attribute = safe_strdup("protocol");
  p->value = safe_strdup("application/pgp-encrypted");

  p->next = mutt_new_parameter();
  p = p->next;
  p->attribute = safe_strdup("boundary");
  p->value = mutt_generate_boundary();

  t->parts = mutt_new_body();
  t->parts->type = TYPEAPPLICATION;
  t->parts->subtype = safe_strdup("pgp-encrypted");
  t->parts->encoding = ENC7BIT;
  t->parts->use_disp = 0;

  t->parts->next = mutt_new_body();
  t->parts->next->type = TYPEAPPLICATION;
  t->parts->next->subtype = safe_strdup("octet-stream");
  t->parts->next->encoding = ENC7BIT;
  t->parts->next->filename = tempfile;
  t->parts->next->use_disp = 0;
  t->parts->next->unlink = 1; /* delete this part after sending the message. */

  mutt_free_body(&a); /* no longer needed! */
  return(t);
}

int pgp_sendMenu (void)
{
  int c = -1;

  ClearLine(LINES-2);
  CenterLine("(e)ncrypt, (s)ign, (b)oth or (f)orget it.", LINES-2);
  ClearLine(LINES-3);
  addstr("pgp: ");
  while (c == -1) {
    refresh();
    switch (c = getch()) {
    case 'e':
      mutt_error("This message will be PGP encrypted.");
      c = PGPENCRYPT;
      break;
    case 's':
      mutt_error("This message will be PGP signed.");
      c = PGPSIGN;
      break;
    case 'b':
      mutt_error("This message will be both PGP signed and encrypted.");
      c = PGPENCRYPT | PGPSIGN;
      break;
    case 'f':
      mutt_error("Warning!  This message will not be signed or encrypted.");
      c = 0;
      break;
    default:
      beep();
      c = -1;
    }
  }
  ClearLine(LINES-2);
  return c;
}
#endif /* _PGPPATH */
