/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

/* Support for Mixmaster remailers. */

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "mutt.h"
#include "mutt_curses.h"
#include "keymap.h"
#include "mutt_menu.h"

#ifdef MIXMASTER_PATH
#define MAXREMAILERS 50

static int Mix_path[20];
static char *Remailers[MAXREMAILERS];

void mix_makeEntry (char *s, int num)
{
  sprintf(s, "%2d   %s", num+1, Remailers[num]);
};

int mix_menu (void)
{
  FILE *p;
  char path[256], *c;
  int max = 0, i, offset = 0, done = 0;
  MUTTMENU *menu;

  for(i=0;i<20;i++) Mix_path[i] = -1;
  sprintf(path, "%s -T", MIXMASTER_PATH);
  p = popen(path, "r");
  while (fgets(path, sizeof(path)-1, p) != NULL) {
    c = strchr(path, ' ');
    *c = 0;
    Remailers[max++] = safe_strdup(path);
  }
  pclose(p);
  Remailers[max] = 0;
  path[0] = 0;

  menu = mutt_newMenu();
  menu->max = max;
  menu->makeEntry = mix_makeEntry;
  menu->menu = MENU_REMAILER;
  menu->title = safe_strdup("Mixmaster Menu");
#ifdef NCURSES_MOUSE_VERSION
  menu->tool = Generic_toolbar;
  toolbar_init(Generic_toolbar);
#endif

  while (!done) {
    switch (mutt_menuLoop(menu)) {

    case OP_GENERIC_SELECT_ENTRY:
    
      Mix_path[offset++] = menu->current;
      if (path[0])
        sprintf(path+strlen(path), ", %s", Remailers[menu->current]);
      else
        strcpy(path, Remailers[menu->current]);
      break;

    case OP_GENERIC_EXIT:

      done = 1;
      break;
    }
  }
  for(i=0; Remailers[i];i++) safe_free((void **)&Remailers[i]);
  mutt_menuDestroy(&menu);
  return(offset > 0 ? offset : -1);
}

int mix_sendMessage (ENVELOPE *e, BODY *a)
{
  char path[256];
  FILE *p;
  int i;

  sprintf(path,"%s -f -l ", MIXMASTER_PATH);
  for (i=0; Mix_path[i] > -1; i++)
    sprintf(path+strlen(path), " %d", Mix_path[i]);
  dprint(2,(debugfile, "mix_sendMessage(): executing: %s\n", path));
  p = popen(path, "w");
  fprintf(p, "%s@%s\n\n", e->to->mailbox, e->to->host);
  if (e->subject) fprintf(p, "Subject: %s\n", e->subject);
  fputs("Mime-Version: 1.0\n", p);
  mutt_write_mime_header(a, p);
  fputc('\n', p); /* terminate the header */

  /*
   * Now write the message to the pipe.  -f assumes that the file is "stdin"
   * instead of a normal file.
   */
  mutt_write_mime_body(a, p);
  pclose(p);

  return(0);
}
#endif /* MIXMASTER_PATH */
