/* Copyright (C) 1996 Park Myeong Seok.
 * Original Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include "muttlib.h"
#include "rfc1522.h"

/* @(#)bq.c:encode/decode modules   is@ev.trigem.co.kr  1992.7.22

Copyright (C) 1992 Ienup Sung.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

static char *b64_alphabet = 
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

/*
** Copyright (C) 1992 Ienup Sung.
**    - simplified by pms@romance.kaist.ac.kr
*/
void DecodeBase64(char *ob, const char *ib)
{
    register unsigned long bm;
    register short cnt;

    for(bm=cnt=0; *ib!=0x00; ib++) {
        bm = (*ib==0x2B)?  (bm << 6) | 0x3E:
             (*ib==0x2F)?  (bm << 6) | 0x3F:
             isdigit(*ib)? (bm << 6) | (0x34 + *ib - 0x30):
             isupper(*ib)? (bm << 6) | (*ib - 0x41):
             islower(*ib)? (bm << 6) | (0x1A + *ib - 0x61):
             (bm << 6) | 0;
        if(++cnt >= 4) {
            *ob++ = (char)((bm & 0x0FF0000) >> 16);
            *ob++ = (char)((bm & 0x000FF00) >> 8);
            *ob++ = (char) (bm & 0x00000FF);
            bm = cnt = 0;
        }
    }
    *ob = 0;
}

/*
** Copyright (C) 1992 Ienup Sung.
**    - simplified by pms@romance.kaist.ac.kr
*/
void EncodeBase64(char *ob, const char *ib)
{
    register unsigned long bm;
    register short int i, cnt;

    while (*ib) {
        for(i = cnt = 0, bm = 0; i < 3; i++) {
            if(*ib) {
                bm = (bm << 8) | (unsigned long)(*ib++ & 0x0FF);
                cnt++;
            } else bm = (bm << 8) | (unsigned long)0;
		}
        *ob++ = b64_alphabet[(0x0FC0000 & bm) >> 18];
        *ob++ = b64_alphabet[(0x003F000 & bm) >> 12];
        *ob++ = b64_alphabet[(cnt>=2)? (0x0000FC0 & bm) >> 6: 64];
        *ob++ = b64_alphabet[(cnt==3)? (0x000003F & bm): 64];
    }
    *ob = 0;
}

int HeaderCode(char c)
{
    return ((c==0x42) | (c==0x62))? 4: ((c==0x51) | (c==0x71))? 3: 0;
}

/*
**  Quoted Printable decoding
**    -pms@romance.kaist.ac.kr
*/
int DeQP(const char *s)
{
    register i, hc, hb;
    
    for(i=hc=0; i<2; i++) {
        hb = (s[i] > 0x2F && s[i] < 0x3A)? s[i] - 0x30:
             (s[i] > 0x40 && s[i] < 0x47)? s[i] - 0x37:
/* ignore lower case character!
             (s[i] > 0x60 && s[i] < 0x67)? s[i] - 0x57:
*/
             -1;
        if(hb<0) return 0;
        hc = (i)? hc | hb: hb << 4; 
    }
    return hc;
}

/*
**  Quoted Printable decoding
**    -pms@romance.kaist.ac.kr
*/
void DecodeQP(char *d, const char *s)
{
    char c, h[2];
    register i, j, sl = strlen(s);

    for(i=0; i<sl; i++) {
        if(s[i]==0x3D && i<sl-2) {
            for(j=0; j<2; j++) h[j] = s[i+1+j];
            if(!(c=DeQP(h))) *d++ = 0x3D;
            else { i+= 2; *d++ = c; }
		}
		else if(s[i] == 0x5F) *d++ = 0x20;
        else *d++ = s[i];
    }
	*d = 0;
}

/*
**  Header B-encoding
**    -pms@romance.kaist.ac.kr
*/
void rfc1522_B_encode_string(char *d, unsigned char *s, int l)
{
    unsigned char *e = d;
	register i, count;

	/* First check to see if there are any 8-bit characters */
	for(i=count=0; i<strlen(s); i++) if (s[i] & 0x80) count++;
	if (!count) {
		strfcpy(d, s, l);
		return;
	}

    while (*s) {
        unsigned char ib[2048]="", *h;
        h = ib;
        while (*s && *s<0x80) *e++ = *s++;
        while (*s && *s>=0x80) *h++ = *s++;
        *e = *h = 0;
        if (ib[0]) {
            unsigned char ob[2048]="";
            EncodeBase64(ob, ib);
            sprintf(e, "=?EUC-KR?B?%s?=", ob);
            e += (strlen(ob) + 13);
        }
    }
    *e = 0;
}

void rfc1522_Q_encode_string (char *d, unsigned char *s, int l)
{
  unsigned char *p = s, c, *pword = 0;
  char *wptr = d, charset[32];
  int charsetlen, count = 0, trailer = 0;

  /* First check to see if there are any 8-bit characters */
  while ((c = *p++)) {
    if (c & 0x80)
      count++;
    else if (count == 0 && c == ' ')
      pword = p;
  }
  if (count == 0) {
    strfcpy(d, s, l);
    return;
  }

  /* Copy the first portion of the string which does not need to be encoded. */
  if (pword) {
    unsigned char savechar;
    short len;

    savechar = *pword;
    *pword = 0;
    strcpy(wptr, s);
    len = strlen(wptr);
    wptr += len;
    l -= len;
    *pword = savechar;
    p = pword;
  }
  else
    p = s;

  /* Generate an appropriate RFC1522 word */
  sprintf(charset, "=?%s?Q?", strcasecmp("US-ASCII", Charset) == 0 ? "UNKNOWN-8BIT" : Charset);
  charsetlen = strlen(charset);

  l -= charsetlen;
  strcpy(wptr, charset);
  wptr += charsetlen;

  while ((c = *p++)) {
    /* Encode all 8-bit characters plus any characters that would cause the
     * string to be enclosed in quotes ("") by the rfc822_*() routines.
     * Encoded words are not legal inside of quotes ("") according to
     * RFC1522.
     */
    if (c & 0x80 || strchr("?=_.[]()\\\"@<>/,;:", c) != 0) {      
      sprintf(wptr, "=%2.2X", c);
      wptr += 3;
    }
    else if (c == ' ') {
      if (count == 0) {
	/* Finished encoding! */
	l -= 3;
	*wptr++ = '?';
	*wptr++ = '=';
	*wptr++ = ' ';
	trailer++;
	break;
      }
      else {
	*wptr++ = '_';
	l--;
      }
    }
    else {
      *wptr++ = c;
      l--;
    }
    
    if (c & 0x80)
      count--;
  }

  if (!trailer) {
    *wptr++ = '?';
    *wptr++ = '=';
    *wptr = 0;
    trailer = 0;
  }
  else
    strcpy(wptr, p);
}

/* PMS */
void rfc1522_encode_string(char *d, unsigned char *s, int l)
{
	if(!strcasecmp(Charset, "euc-kr")||!strcasecmp(Charset, "iso-2022-kr"))
		rfc1522_B_encode_string(d, s, l);
	else rfc1522_Q_encode_string(d, s, l);
}

void rfc1522_encode_address (ADDRESS *addr)
{
  ADDRESS *ptr = addr;
  char buffer[STRING];

  while (ptr) {
    if (ptr->personal) {
      rfc1522_encode_string (buffer, (unsigned char *)&ptr->personal, sizeof(buffer));
      safe_free((void **)&ptr->personal);
      ptr->personal = safe_strdup(buffer);
    }
    ptr = ptr->next;
  }
}

static int rfc1522_decode_word (char *d, const char *s, int len)
{
  char *p = safe_strdup(s);
  char *pp = p;
  char *pd = d;
  int enc = 0, filter = 0, count = 0;

  while ((pp = strtok (pp, "?")) != NULL) {
    count++;
    switch (count) {
    case 2:
/* PMS - COMPATCHAR */
      if (strcasecmp (pp, Charset) != 0) {
	if (compat_charset(Charset) && compat_charset(pp)) {
	/* Charset doesn't match, but it's a superset of US-ASCII */
	  filter = 1;
	}
	else {
	/* Can't display this charset! */
	  strncpy(d, s, len);
	  return(0);
	}
      }
      break;
    case 3:
      if (toupper(*pp) == 'Q') enc = ENCQUOTEDPRINTABLE;
      else if (toupper(*pp) == 'B') enc = ENCBASE64;
      else return (-1);
      break;
    case 4:
      if (enc == ENCQUOTEDPRINTABLE) DecodeQP(pd, pp);
      else if (enc == ENCBASE64) DecodeBase64(pd, pp);
      break;
    }
    pp = 0;
  }
  safe_free ((void **)&p);
  if (filter) {
    pd = d;
    while (*pd) {
      if (*pd & 0x80 || !isprint(*pd)) *pd = '?';
      pd++;
    }
  }
  return (0);
}

void rfc1522_decode_string (char *d, const char *s, int len)
{
  char *q = safe_strdup(s);
  char *p = q;
  char *pp = p;
  char *pd = d;
  char savechar = 0;
  int toklen;

  while (*p && len > 0) {
    pp = strpbrk(p, " \t\n");
    if (pp) {
      savechar = *pp;
      *pp = 0;
      toklen = pp - p;
    }
    else {
      toklen = strlen (p);
      savechar = 0;
    }

    if (p[0] == '=' && p[1] == '?' && p[toklen-2] == '?' && p[toklen-1] == '=')
      rfc1522_decode_word(pd, p, len);
    else
      strncpy(pd, p, len);

    len -= strlen(pd);
    pd += strlen(pd);

    if (savechar) {
      *pd++ = savechar;
      len--;
      savechar = 0;
    }

    if (pp)
      p = pp + 1;
    else
      break;
  }
  *pd = 0;
  safe_free((void **)&q);
}

/* Gack!  Parsing RFC822 structured fields is a royal pain... */
static int rfc822_toklen (char *s)
{
  char *p = s;
  short depth;

  while (p) {
    if (*p == '(') {
      depth = 0;
      while (*p) {
        if (*p == '\\')
	  p++;
        else if (*p == '\"')
	  p += rfc822_toklen(p);
        else if (*p == '(')
	  depth++;
        else if (*p == ')')
	  depth--;
        p++;
        if (depth == 0)
	  break;
      }
      break;
    }
    else if (*p == '\"') {
      p++;
      while (*p) {
        if (*p == '\\')
	  p++;
        else if (*p == '\"')
	  break;
	p++;
      }
      break;
    }
    else if (strchr(".:@ \t", *p) != 0)
      break;
    p++;
  }
  return(p != s ? p-s : 1);
}

static int comment_toklen (char *s)
{
  char *p = strpbrk(s, " \r\n\t(");
  if (p) return (p != s ? p-s : 1);
  else return(strlen(s));
}

#define IN_STRUCTURED 1
#define IN_COMMENT    2
#define IN_PHRASE     4

int rfc1522_decode_etc (char *d, char *s, int len, int flag)
{
  int toklen=1, wordlen;
  char *ps, savechar, *pd = d;
  
  while (*s && (len > 0)) {
    if (flag & IN_STRUCTURED)
      toklen = rfc822_toklen(s);
    else if (flag & IN_COMMENT)
      toklen = comment_toklen(s);
    
    ps = s + toklen;
    savechar = *ps;
    *ps = 0;
    
    if (((flag & IN_STRUCTURED) || (flag & IN_COMMENT)) && *s == '(') {
      *pd++ = '(';
      len--;
      s[toklen-1] = 0;
      if (rfc1522_decode_etc(pd, s+1, len, IN_COMMENT) == -1) return(-1);
      wordlen = strlen(pd);
      pd += wordlen;
      len -= wordlen;
      *pd++ = ')';
      len--;
    }
    else if ((((flag & IN_STRUCTURED) && (flag & IN_PHRASE)) ||
        !(flag & IN_STRUCTURED)) &&
       (s[0] == '=' && s[1] == '?' && s[toklen-2] == '?' && s[toklen-1] == '=')) {
      if (rfc1522_decode_word(pd, s, len) == -1) return(-1);
      wordlen = strlen(pd);
      pd += wordlen;
      len -= wordlen;
    }
    else {
      len -= toklen;
      if (len < 0) return(-1);
      strncpy(pd, s, toklen);
      pd += toklen;
    }

    *ps = savechar;
    s = ps;
  }
  *pd = 0;
  return(0);
}

/*
**  Email Header Decoding
**    - pms@romance.kaist.ac.kr
**      QP=3, B64=4
*/
int DeHeader(char *d, int code, const char *s)
{
    char ts[1024];
	int i, len;

    if(code==3) DecodeQP(ts, s);
    if(code==4) DecodeBase64(ts, s);
	len = strlen(ts);
	for(i=0; i<len; i++) *d++ = ts[i];
	return len;
}

/*
**  Email Header Decoding
**    - pms@romance.kaist.ac.kr
**      QP=3, B64=4
*/
void DecodeHeader(char *d, const char *s)
{
    char ts[LONG_STRING];
    register i, j, c, code, sl, fail;

    sl = strlen(s);
    for(i=c=code=0; i<sl; i++) {
        if(i<sl-1 && s[i]==0x3D && s[i+1]==0x3F) {
            fail = 0;
            if(i<sl-3 && s[i+3]==0x3F) {
                if(!code) code = HeaderCode(s[i+2]);
                if(code) {
                    for(j=0;j<sl-i-4 && s[i+4+j]!=0x3F;j++);
                    if(j>sl-i-4) fail = 1;
                    else c = i + 5 + j;
                } else fail = 1;
            }
            else {
                for(j=0; j<sl-i-2 && s[i+2+j]!=0x3F; j++);
                if(j<sl-i-3) {
                    if(!code) code = HeaderCode(s[i+3+j]); 
                    if(code && j<sl-i-4 && s[i+4+j]==0x3F) c = i + 5 + j;
                    else fail = 1;
                } else fail = 1;
            }
            if(!fail) {
                for(j=0;c<sl && s[c]!=0x3F; j++, c++) ts[j] = s[c];
                if(c<sl-1 && s[c+1]==0x3D) {
                    c++;
                    ts[j] = 0; d += DeHeader(d, code, ts);
                } else fail = 1;
            }
            if(fail) { 
                for(j=0; j<2; j++) *d++ = s[i+j];
                i++;
            } else i = c;
        } else *d++ = s[i]; 
    }
	*d = 0;
}

int rfc1522_decode_real (char *d, char *s, int len, int flag)
{
    if ((flag & IN_STRUCTURED) || (flag & IN_COMMENT))
		return rfc1522_decode_etc(d, s, len, flag);
    else DecodeHeader(d, s);
	return 0;
}

int rfc1522_valid (const char *s)
{
  char *p = strstr(s, "=?");
  if (!p) return(0);
  p = strstr(p+2, "?=");
  if (!p) return(0);
  return(1);
}

#define HDR_STRUCTURED 1
#define HDR_PHRASE     2

void rfc1522_decode (char *s, int len, int flags)
{
  char buffer[LONG_STRING];
  int passflags = 0;

  if (!rfc1522_valid(s)) return; /* nothing to do */
  strncpy(buffer, s, sizeof(buffer));
  if (flags & HDR_STRUCTURED) {
    passflags = IN_STRUCTURED;
    if (flags & HDR_PHRASE)
      passflags |= IN_PHRASE;
  }
  if (rfc1522_decode_real(s, buffer, len, passflags) == -1)
    strcpy(s, buffer);
}

/*
 * This piece of code, and the idea of using the HDR_* flags, is attributed
 * to Kari Hurtta <kari.hurtta@fmi.fi>, who helped me get this right.
 */
int rfc1522_classify (const char *s)
{
  if (strncasecmp("from:", s, sizeof("from:")-1) == 0 ||
      strncasecmp("to:", s, sizeof("to:")-1) == 0 ||
      strncasecmp("cc:", s, sizeof("cc:")-1) == 0 ||
      strncasecmp("bcc:", s, sizeof("bcc:")-1) == 0)
    return(HDR_STRUCTURED | HDR_PHRASE);
  return(0);
}

void rfc1522_decode_adrlist (ADDRESS *a)
{
  while (a) {
    if (a->personal && rfc1522_valid(a->personal)) {
      rfc1522_decode(a->personal, strlen(a->personal), 0);
      safe_realloc((void **)&a->personal, strlen(a->personal)+1);
    }
    a = a->next;
  }
}

void rfc1522_encode_adrlist (ADDRESS *p)
{
  char buffer[LONG_STRING];

  while (p) {
    if (p->personal) {
      rfc1522_encode_string(buffer, p->personal, sizeof(buffer)-1);
      safe_free((void **)&p->personal);
      p->personal = safe_strdup(buffer);
    }
    p = p->next;
  }
}
