/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <errno.h>
#include <unistd.h>
#include <string.h>
#include "mutt.h"
#include "mutt_curses.h"

#define ATT_REDRAW_FULL (1<<1)
#define ATT_REDRAW_INDEX (1<<2)
#define ATT_REDRAW_INFO (1<<3)
#define ATT_OFFSET 2
#define ATT_XOFFSET 27 /* where the data for the content- headers begins */
#define ATT_TYPE (LINES-9)
#define ATT_ENCODING (LINES-8)
#define ATT_DESCRIPTION (LINES-7)
#define ATT_HELP (LINES-5)
#define ATT_LENGTH (ATT_TYPE-ATT_OFFSET-1)

BODY *ci_attach_menu (BODY *att)
{
  BODY **iindex = 0, *ptr = att;
  int redraw = ATT_REDRAW_FULL;
  int iindexlen = 0, count = 0, cur = 1, top = 1, i, ch;
  char buffer[STRING];

  while (ptr) {
    if (iindexlen == count) {
      if (iindexlen == 0)
        iindex = (BODY **)safe_malloc(sizeof(BODY *)*((iindexlen += 5)));
      else
        safe_realloc((void **)&iindex, sizeof(BODY *)*((iindexlen += 5)));
    }
    iindex[count++] = ptr;
    ptr = ptr->next;
  }

  FOREVER {
    if (redraw & ATT_REDRAW_FULL) {
      clear();
      CenterLine("Attachment Menu", 0);
      CenterLine("(a)dd, (d)elete, or (q)uit.", ATT_HELP);
      redraw = ATT_REDRAW_INDEX | ATT_REDRAW_INFO;
    }

    if (redraw & ATT_REDRAW_INDEX) {
      for (i=top; i<ATT_LENGTH; i++) {
        if (i <= count) {
          sprintf(buffer, "  %2d  %-*.*s", i, COLS-6, COLS-6, iindex[i-1]->filename);
          move(i-top+ATT_OFFSET, 0);
          if (i == cur) attrset(A_REVERSE);
          addstr(buffer);
          if (i == cur) attrset(A_NORMAL);
        }
        else
          ClearLine(i-top+ATT_OFFSET);
      }
      redraw &= ~ATT_REDRAW_INDEX;
    }

    if (redraw & ATT_REDRAW_INFO) {
      if (count > 0)
        sprintf(buffer, "%s/%s", TYPE(iindex[cur-1]->type), iindex[cur-1]->subtype);
      else
        buffer[0] = 0;

      ClearLine(ATT_TYPE);
      printw("Content-type:              %-*.*s",
        COLS-ATT_XOFFSET, COLS-ATT_XOFFSET, buffer);
      ClearLine(ATT_ENCODING);
      printw("Content-transfer-encoding: %-*.*s",
        COLS-ATT_XOFFSET, COLS-ATT_XOFFSET,
         count > 0 ? ENCODING(iindex[cur-1]->encoding) : "");
      ClearLine(ATT_DESCRIPTION);
      printw("Content-de(S)cription:     %-*.*s",
        COLS-ATT_XOFFSET, COLS-ATT_XOFFSET,
         count > 0 ? (iindex[cur-1]->description ? iindex[cur-1]->description : "") : "");
      redraw &= ~ATT_REDRAW_INFO;
    }

    move(cur-top+ATT_OFFSET, COLS-1);
    refresh();
    ch = getch();
    ci_clear_error();

    if (ch == 'j' || ch == KEY_DOWN) {
      if (cur < count) {
        cur++;
        redraw = ATT_REDRAW_INDEX | ATT_REDRAW_INFO;
      }
      else
        mutt_error("You are on the last attachment.");
    }
    else if (ch == 'k' || ch == KEY_UP) {
      if (cur > 1) {
        cur--;
        redraw = ATT_REDRAW_INDEX | ATT_REDRAW_INFO;
      }
      else
        mutt_error("You are on the first attachment.");
    }
    else if (ch == 'a') {
      buffer[0] = 0;
      if (ci_get_field("Attach file: ", buffer, 0, CI_complete) == -1 || buffer[0] == 0)
	continue;
      mutt_expand_path(buffer);
      if (access(buffer, F_OK|R_OK) == -1) {
        if (errno == ENOENT)
          mutt_error("That file does not seem to exist!");
        else if (errno == EACCES)
          mutt_error("You don't have permission to read that file!");
        else
          mutt_error("Could not access file: errno=%d", errno);
        continue;
      }
      if (iindexlen == count) {
        if (iindexlen == 0)
          iindex = (BODY **)safe_malloc(sizeof(BODY *) * (iindexlen += 5) );
        else
          safe_realloc((void **)&iindex, sizeof(BODY *) * (iindexlen += 5));
      }
      iindex[count] = mutt_new_body();
      iindex[count]->filename = safe_strdup(buffer);
      if (count > 0) iindex[count-1]->next = iindex[count];
      iindex[count]->content = mutt_get_content_info(buffer);
      mutt_set_content_info(iindex[count]);
      count++;
      cur = count;
      redraw = ATT_REDRAW_INDEX | ATT_REDRAW_INFO;
    }
    else if (ch == 'd') {
      if (count) {
        if (cur == count) {
          cur--;
          count--;
          mutt_free_body(&iindex[count]);
          if (count > 0) iindex[count-1]->next = 0;
        }
        else {
          /* Don't clobber the rest of the attachments! */
          iindex[cur-1]->next = 0;
          mutt_free_body(&iindex[cur-1]);
          for (i=cur-1; i<count-1; i++)
            iindex[i] = iindex[i+1];
          iindex[count-1] = 0;
          count--;
          for (i=1; i<=count; i++)
            iindex[i-1]->next = iindex[i];
        }
        redraw = ATT_REDRAW_INDEX | ATT_REDRAW_INFO;
      }
      else
        mutt_error("There are no attachments.");
    }
    else if (ch == 's') {
      if (iindex[cur-1]->description)
        strcpy(buffer, iindex[cur-1]->description);
      else
        buffer[0] = 0;
      if (ci_enter_string(buffer, ATT_DESCRIPTION, ATT_XOFFSET, 0, 0) != -1 && buffer[0] != 0) {
        safe_free((void **)&iindex[cur-1]->description);
        iindex[cur-1]->description = safe_strdup(buffer);
      }
      if (buffer[0] == 0 && iindex[cur-1]->description)
        safe_free((void **)&iindex[cur-1]->description);
    }
    else if (ch == 'q')
      break;
    else if (ch == ctrl('L'))
      redraw = ATT_REDRAW_FULL;
    else 
      beep();
  }
  if (iindex) {
    att = iindex[0];
    safe_free((void **)&iindex);
  }
  return(att);
}
