/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <curses.h>
#include <signal.h>
#include <setjmp.h>
#include "mutt.h"

extern int ci_sigwinch;

static int ci_isendwin = 0;

/* Attempt to catch "ordinary" signals and shut down gracefully. */
RETSIGTYPE ci_sighandler (int sig)
{
  endwin(); /* just to be safe */
#if SYS_SIGLIST_DECLARED
  printf("Caught %s...  Exiting.\n", sys_siglist[sig]);
#else
#if (__sun__ && __svr4__)
  printf("Caught %s...  Exiting.\n", _sys_siglist[sig]);
#else
  printf("Caught signal %d...  Exiting.\n", sig);
#endif
#endif
  exit(0);
}

RETSIGTYPE ci_signal_suspend (int d)
{
  signal(SIGTSTP, SIG_DFL);
  ci_isendwin = isendwin();
  endwin(); /* just to be safe */
  puts("Suspended.  Type \"fg\" to resume.");
  fflush(stdout);
  kill(0, SIGSTOP);
}

RETSIGTYPE ci_signal_resume (int d)
{
  signal(SIGCONT, ci_signal_resume);
  signal(SIGTSTP, ci_signal_suspend);
  if (!ci_isendwin) refresh();
}

RETSIGTYPE ci_resize_screen (int d)
{
  ci_sigwinch = 1;
  signal(SIGWINCH, ci_resize_screen);
  longjmp(Jumpbuf, 1);
}

void ci_signal_init (void)
{
  signal(SIGTERM, ci_sighandler);
  signal(SIGHUP, ci_sighandler);
  signal(SIGINT, ci_sighandler);
  signal(SIGTSTP, ci_signal_suspend);
  signal(SIGCONT, ci_signal_resume);
  signal(SIGWINCH, ci_resize_screen);
  signal(SIGPIPE, SIG_IGN);
}
